from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_preset import V1Preset


class V1ListPresetsResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[V1Preset]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from typing import Any, Dict, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1Tag(BaseResponseModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    color: Optional[StrictStr]
    description: Optional[StrictStr]
    icon: Optional[StrictStr]
    stats: Optional[Dict[str, Any]]



from typing import Optional

from pydantic import StrictStr

from polyaxon import dist
from polyaxon.schemas.api.compatibility import V1Compatibility
from polyaxon.schemas.api.installation import V1Installation
from polyaxon.schemas.api.log_handler import V1LogHandler
from polyaxon.schemas.cli.checks_config import ChecksConfig


class CliConfig(ChecksConfig):
    _IDENTIFIER = "cli"
    _DIST = "dist"
    _INTERVAL = 30 * 60

    current_version: Optional[StrictStr]
    installation: Optional[V1Installation]
    compatibility: Optional[V1Compatibility]
    log_handler: Optional[V1LogHandler]

    @property
    def min_version(self) -> Optional[str]:
        if not self.compatibility or not self.compatibility.cli:
            return None
        return self.compatibility.cli.min

    @property
    def latest_version(self) -> Optional[str]:
        if not self.compatibility or not self.compatibility.cli:
            return None
        return self.compatibility.cli.latest

    @property
    def is_community(self) -> bool:
        return self.installation is None or dist.is_community(self.installation.dist)



from typing import Dict, List, Optional

from clipped.utils.lists import to_list

from polyaxon.connections import V1Connection
from polyaxon.polyflow import V1IO, V1CompiledOperation
from polyaxon.types import IMAGE, LINEAGE_VALUES
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def collect_artifacts_from_io(
    io: V1IO, connection_by_names: Dict[str, V1Connection], is_input: bool
) -> Optional[V1RunArtifact]:
    if io.type not in LINEAGE_VALUES:
        return None

    if io.type == IMAGE:
        image = io.value
        connection = connection_by_names.get(io.connection)
        if connection and connection.schema_ and connection.schema_.url:
            image = "{}/{}".format(connection.schema_.url, image)
        return V1RunArtifact(
            name=io.name,
            kind=V1ArtifactKind.DOCKER_IMAGE,
            connection=io.connection,
            summary={"image": image},
            is_input=is_input,
        )


def collect_artifacts_from_io_section(
    io_section: List[V1IO],
    connection_by_names: Dict[str, V1Connection],
    is_input: bool,
) -> List[V1RunArtifact]:
    io_section = to_list(io_section, check_none=True)
    artifacts = [
        collect_artifacts_from_io(
            io, connection_by_names=connection_by_names, is_input=is_input
        )
        for io in io_section
    ]
    return [a for a in artifacts if a]


def collect_io_artifacts(
    compiled_operation: V1CompiledOperation,
    connection_by_names: Dict[str, V1Connection],
) -> List[V1RunArtifact]:
    connection_by_names = connection_by_names or {}
    artifacts = []
    artifacts += collect_artifacts_from_io_section(
        compiled_operation.inputs,
        connection_by_names=connection_by_names,
        is_input=True,
    )
    artifacts += collect_artifacts_from_io_section(
        compiled_operation.outputs,
        connection_by_names=connection_by_names,
        is_input=False,
    )
    return artifacts



import uuid

from polyaxon.runner.converter.common import constants


def get_volume_name(path: str) -> str:
    name = uuid.uuid5(namespace=uuid.NAMESPACE_DNS, name=path).hex
    return constants.VOLUME_MOUNT_CONNECTIONS_FORMAT.format(name)



from typing import Optional

from pydantic import Extra, Field, StrictStr

from polyaxon.env_vars.keys import EV_KEYS_AUTH_TOKEN, EV_KEYS_AUTH_USERNAME
from polyaxon.schemas.base import BaseSchemaModel


class AccessTokenConfig(BaseSchemaModel):
    """
    Access token config.


    Args:
        username: `str`. The user's username.
        token: `str`. The user's token.
    """

    _IDENTIFIER = "token"

    username: Optional[StrictStr] = Field(alias=EV_KEYS_AUTH_USERNAME)
    token: Optional[StrictStr] = Field(alias=EV_KEYS_AUTH_TOKEN)

    class Config:
        extra = Extra.ignore


class V1Credentials(BaseSchemaModel):
    """
    Credentials config.


    Args:
        username: `str`. The user's username.
        password: `str`. The user's password.
    """

    _IDENTIFIER = "credentials"

    username: StrictStr
    password: StrictStr



from typing_extensions import Literal

from clipped.types.ref_or_obj import DatetimeOrRef
from pydantic import Field

from polyaxon.polyflow.schedules.kinds import V1ScheduleKind
from polyaxon.schemas.base import BaseSchemaModel


class V1DateTimeSchedule(BaseSchemaModel):
    """Date schedule is an interface to kick a component execution at a specific time.

    Args:
        kind: str, should be equal to `datetime`
        start_at: datetime, required

    ## YAML usage

    ```yaml
    >>> schedule:
    >>>   kind:
    >>>   startAt:
    ```

    ## Python usage

    ```python
    >>> from datetime import datetime
    >>> from polyaxon.polyflow import V1DateTimeSchedule
    >>> schedule = V1DateTimeSchedule(
    >>>   start_at=datetime(...),
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this schedule
    is a datetime schedule.

    If you are using the python client to create the schedule,
    this field is not required and is set by default.

    ```yaml
    >>> run:
    >>>   kind: datetime
    ```

    ### startAt

    Optional field to set the start time for kicking the first execution,
    all following executions will be relative to this time.

    ```yaml
    >>> run:
    >>>   startAt: "2019-06-24T21:20:07+00:00"
    ```
    """

    _IDENTIFIER = V1ScheduleKind.DATETIME
    _USE_DISCRIMINATOR = True

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    start_at: DatetimeOrRef = Field(alias="startAt")

    @property
    def max_runs(self):
        return 0

    @property
    def end_at(self):
        return None

    @property
    def depends_on_past(self):
        return True



from typing import Optional, Union
from typing_extensions import Literal

from clipped.types.ref_or_obj import RefField
from pydantic import Field

from polyaxon.k8s.k8s_schemas import V1Container
from polyaxon.polyflow.run.base import BaseRun
from polyaxon.polyflow.run.kinds import V1RunKind
from polyaxon.polyflow.run.kubeflow.clean_pod_policy import V1CleanPodPolicy
from polyaxon.polyflow.run.kubeflow.replica import V1KFReplica
from polyaxon.polyflow.run.kubeflow.scheduling_policy import V1SchedulingPolicy
from polyaxon.polyflow.run.resources import V1RunResources
from polyaxon.polyflow.run.utils import DestinationImageMixin


class V1TFJob(BaseRun, DestinationImageMixin):
    """Kubeflow TF-Job provides an interface to train distributed experiments with TensorFlow.

    Args:
        kind: str, should be equal `tfjob`
        clean_pod_policy: str, one of [`All`, `Running`, `None`]
        scheduling_policy: [V1SchedulingPolicy](/docs/experimentation/distributed/scheduling-policy/), optional  # noqa
        enable_dynamic_worker: boolean
        chief: [V1KFReplica](/docs/experimentation/distributed/kubeflow-replica/), optional
        ps: [V1KFReplica](/docs/experimentation/distributed/kubeflow-replica/), optional
        worker: [V1KFReplica](/docs/experimentation/distributed/kubeflow-replica/), optional
        evaluator: [V1KFReplica](/docs/experimentation/distributed/kubeflow-replica/), optional

    ## YAML usage

    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   cleanPodPolicy:
    >>>   schedulingPolicy:
    >>>   enableDynamicWorker:
    >>>   chief:
    >>>   ps:
    >>>   worker:
    >>>   evaluator:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.polyflow import V1KFReplica, V1TFJob
    >>> from polyaxon.k8s import k8s_schemas
    >>> tf_job = V1TFJob(
    >>>     clean_pod_policy='All',
    >>>     chief=V1KFReplica(...),
    >>>     ps=V1KFReplica(...),
    >>>     worker=V1KFReplica(...),
    >>>     evaluator=V1KFReplica(...),
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this component's runtime is a tfjob.

    If you are using the python client to create the runtime,
    this field is not required and is set by default.

    ```yaml
    >>> run:
    >>>   kind: tfjob
    ```

    ### cleanPodPolicy

    Controls the deletion of pods when a job terminates.
    The policy can be one of the following values: [`All`, `Running`, `None`]


    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   cleanPodPolicy: 'All'
    >>>  ...
    ```

    ### schedulingPolicy

    SchedulingPolicy encapsulates various scheduling policies of the distributed training
    job, for example `minAvailable` for gang-scheduling.


    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   schedulingPolicy:
    >>>     ...
    >>>  ...
    ```

    ### enableDynamicWorker

    Flag to enable dynamic worker.


    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   enableDynamicWorker: true
    >>>     ...
    >>>  ...
    ```

    ### chief

    The chief is responsible for orchestrating training and performing
    tasks like checkpointing the model.

    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   chief:
    >>>     replicas: 1
    >>>     container:
    >>>       ...
    >>>  ...
    ```

    ### ps

    The ps are parameter servers; these servers provide a distributed data store
    for the model parameters.

    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   ps:
    >>>     replicas: 2
    >>>     container:
    >>>       ...
    >>>  ...
    ```

    ### worker

    The workers do the actual work of training the model. In some cases,
    worker 0 might also act as the chief.

    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   worker:
    >>>     replicas: 2
    >>>     container:
    >>>       ...
    >>>  ...
    ```

    ### evaluator

    The evaluators can be used to compute evaluation metrics as the model is trained.

    ```yaml
    >>> run:
    >>>   kind: tfjob
    >>>   evaluator:
    >>>     replicas: 1
    >>>     container:
    >>>       ...
    >>>  ...
    ```
    """

    _IDENTIFIER = V1RunKind.TFJOB

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    clean_pod_policy: Optional[V1CleanPodPolicy] = Field(alias="cleanPodPolicy")
    enable_dynamic_worker: Optional[bool] = Field(alias="enableDynamicWorker")
    scheduling_policy: Optional[V1SchedulingPolicy] = Field(alias="schedulingPolicy")
    chief: Optional[Union[V1KFReplica, RefField]]
    ps: Optional[Union[V1KFReplica, RefField]]
    worker: Optional[Union[V1KFReplica, RefField]]
    evaluator: Optional[Union[V1KFReplica, RefField]]

    def apply_image_destination(self, image: str):
        if self.chief:
            self.chief.container = self.chief.container or V1Container()
            self.chief.container.image = image
        if self.ps:
            self.ps.container = self.ps.container or V1Container()
            self.ps.container.image = image
        if self.worker:
            self.worker.container = self.worker.container or V1Container()
            self.worker.container.image = image
        if self.evaluator:
            self.evaluator.container = self.evaluator.container or V1Container()
            self.evaluator.container.image = image

    def get_resources(self):
        resources = V1RunResources()
        if self.chief:
            resources += self.chief.get_resources()
        if self.ps:
            resources += self.ps.get_resources()
        if self.worker:
            resources += self.worker.get_resources()
        if self.evaluator:
            resources += self.evaluator.get_resources()
        return resources

    def get_all_containers(self):
        containers = []
        if self.chief:
            containers += self.chief.get_all_containers()
        if self.ps:
            containers += self.ps.get_all_containers()
        if self.worker:
            containers += self.worker.get_all_containers()
        if self.evaluator:
            containers += self.evaluator.get_all_containers()
        return containers

    def get_all_connections(self):
        connections = []
        if self.chief:
            connections += self.chief.get_all_connections()
        if self.ps:
            connections += self.ps.get_all_connections()
        if self.worker:
            connections += self.worker.get_all_connections()
        if self.evaluator:
            connections += self.evaluator.get_all_connections()
        return connections

    def get_all_init(self):
        init = []
        if self.chief:
            init += self.chief.get_all_init()
        if self.ps:
            init += self.ps.get_all_init()
        if self.worker:
            init += self.worker.get_all_init()
        if self.evaluator:
            init += self.evaluator.get_all_init()
        return init



import datetime

from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_project_settings import V1ProjectSettings


class V1Project(BaseResponseModel):
    uuid: Optional[UUIDStr]
    owner: Optional[StrictStr]
    name: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    is_public: Optional[bool]
    bookmarked: Optional[bool]
    readme: Optional[StrictStr]
    excluded_features: Optional[List[StrictStr]]
    excluded_runtimes: Optional[List[StrictStr]]
    settings: Optional[V1ProjectSettings]
    role: Optional[StrictStr]
    live_state: Optional[int]



from typing import Any, Iterable, List, Optional

from clipped.utils.enums import get_enum_value
from clipped.utils.json import orjson_dumps

from polyaxon import settings
from polyaxon.api import VERSION_V1
from polyaxon.connections import V1ConnectionResource
from polyaxon.env_vars.keys import (
    EV_KEYS_API_VERSION,
    EV_KEYS_AUTH_TOKEN,
    EV_KEYS_AUTHENTICATION_TYPE,
    EV_KEYS_HEADER,
    EV_KEYS_HEADER_SERVICE,
    EV_KEYS_HOST,
    EV_KEYS_IS_MANAGED,
    EV_KEYS_K8S_NAMESPACE,
    EV_KEYS_K8S_NODE_NAME,
    EV_KEYS_K8S_POD_ID,
    EV_KEYS_LOG_LEVEL,
    EV_KEYS_SECRET_INTERNAL_TOKEN,
    EV_KEYS_SECRET_KEY,
)
from polyaxon.exceptions import PolyaxonConverterError
from polyaxon.k8s import k8s_schemas
from polyaxon.k8s.converter.common.accelerators import requests_gpu
from polyaxon.runner.converter import BaseConverter
from polyaxon.services.headers import PolyaxonServiceHeaders


class EnvMixin(BaseConverter):
    @staticmethod
    def _get_env_var(name: str, value: Any) -> k8s_schemas.V1EnvVar:
        if not isinstance(value, str):
            try:
                value = orjson_dumps(value)
            except (ValueError, TypeError) as e:
                raise PolyaxonConverterError(e)

        return k8s_schemas.V1EnvVar(name=name, value=value)

    @staticmethod
    def _get_resources_env_vars(
        resources: k8s_schemas.V1ResourceRequirements,
    ) -> List[k8s_schemas.V1EnvVar]:
        env_vars = []
        has_gpu = requests_gpu(resources)

        # Fix https://github.com/kubernetes/kubernetes/issues/59629
        # When resources.gpu.limits is not set or set to 0, we explicitly
        # pass NVIDIA_VISIBLE_DEVICES=none into container to avoid exposing GPUs.
        if not has_gpu:
            env_vars.append(
                k8s_schemas.V1EnvVar(name="NVIDIA_VISIBLE_DEVICES", value="none")
            )

        return env_vars

    @staticmethod
    def _get_from_config_map(
        key_name: str, config_map_key_name: str, config_map_ref_name: str
    ) -> k8s_schemas.V1EnvVar:
        config_map_key_ref = k8s_schemas.V1ConfigMapKeySelector(
            name=config_map_ref_name, key=config_map_key_name
        )
        value_from = k8s_schemas.V1EnvVarSource(config_map_key_ref=config_map_key_ref)
        return k8s_schemas.V1EnvVar(name=key_name, value_from=value_from)

    @staticmethod
    def _get_from_secret(
        key_name: str, secret_key_name: str, secret_ref_name: str
    ) -> k8s_schemas.V1EnvVar:
        secret_key_ref = k8s_schemas.V1SecretKeySelector(
            name=secret_ref_name, key=secret_key_name
        )
        value_from = k8s_schemas.V1EnvVarSource(secret_key_ref=secret_key_ref)
        return k8s_schemas.V1EnvVar(name=key_name, value_from=value_from)

    @classmethod
    def _get_items_from_secret(
        cls, secret: V1ConnectionResource
    ) -> List[k8s_schemas.V1EnvVar]:
        items_from = []
        if not secret or not secret.items or secret.mount_path:
            return items_from

        for item in secret.items:
            items_from.append(
                cls._get_from_secret(
                    key_name=item, secret_key_name=item, secret_ref_name=secret.name
                )
            )
        return items_from

    @classmethod
    def _get_items_from_config_map(
        cls,
        config_map: V1ConnectionResource,
    ) -> List[k8s_schemas.V1EnvVar]:
        items_from = []
        if not config_map or not config_map.items:
            return items_from

        for item in config_map.items:
            items_from.append(
                cls._get_from_config_map(
                    key_name=item,
                    config_map_key_name=item,
                    config_map_ref_name=config_map.name,
                )
            )
        return items_from

    @staticmethod
    def _get_from_field_ref(name: str, field_path: str) -> k8s_schemas.V1EnvVar:
        field_ref = k8s_schemas.V1ObjectFieldSelector(field_path=field_path)
        value_from = k8s_schemas.V1EnvVarSource(field_ref=field_ref)
        return k8s_schemas.V1EnvVar(name=name, value_from=value_from)

    @classmethod
    def _get_env_vars_from_k8s_resources(
        cls,
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[k8s_schemas.V1EnvVar]:
        secrets = secrets or []
        config_maps = config_maps or []

        env_vars = []
        for secret in secrets:
            env_vars += cls._get_items_from_secret(secret=secret)
        for config_map in config_maps:
            env_vars += cls._get_items_from_config_map(config_map=config_map)

        return env_vars

    @classmethod
    def _get_additional_env_vars(cls) -> List[k8s_schemas.V1EnvVar]:
        return []

    @staticmethod
    def _get_env_from_secret(
        secret: V1ConnectionResource,
    ) -> Optional[k8s_schemas.V1EnvFromSource]:
        if not secret or secret.items or secret.mount_path:
            return None

        return k8s_schemas.V1EnvFromSource(secret_ref={"name": secret.name})

    @classmethod
    def _get_env_from_secrets(
        cls,
        secrets: Iterable[V1ConnectionResource],
    ) -> List[k8s_schemas.V1EnvFromSource]:
        secrets = secrets or []
        results = [cls._get_env_from_secret(secret=secret) for secret in secrets]
        return [r for r in results if r]

    @staticmethod
    def _get_env_from_config_map(
        config_map: V1ConnectionResource,
    ) -> Optional[k8s_schemas.V1EnvFromSource]:
        if not config_map or config_map.items or config_map.mount_path:
            return None

        return k8s_schemas.V1EnvFromSource(config_map_ref={"name": config_map.name})

    @classmethod
    def _get_env_from_config_maps(
        cls,
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[k8s_schemas.V1EnvFromSource]:
        config_maps = config_maps or []
        results = [
            cls._get_env_from_config_map(config_map=config_map)
            for config_map in config_maps
        ]
        return [r for r in results if r]

    @classmethod
    def _get_env_from_k8s_resources(
        cls,
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[k8s_schemas.V1EnvFromSource]:
        secrets = secrets or []
        config_maps = config_maps or []

        env_vars = []
        env_vars += cls._get_env_from_secrets(secrets=secrets)
        env_vars += cls._get_env_from_config_maps(config_maps=config_maps)
        return env_vars

    @classmethod
    def _get_base_env_vars(
        cls,
        namespace: str,
        resource_name: str,
        use_proxy_env_vars_use_in_ops: bool,
        log_level: Optional[str] = None,
    ):
        env = [
            cls._get_from_field_ref(
                name=EV_KEYS_K8S_NODE_NAME, field_path="spec.nodeName"
            ),
            cls._get_from_field_ref(
                name=EV_KEYS_K8S_NAMESPACE, field_path="metadata.namespace"
            ),
            cls._get_from_field_ref(
                name=EV_KEYS_K8S_POD_ID, field_path="metadata.name"
            ),
        ]
        if log_level:
            env.append(cls._get_env_var(name=EV_KEYS_LOG_LEVEL, value=log_level))
        env += cls._get_proxy_env_vars(use_proxy_env_vars_use_in_ops)
        return env

    def _get_service_env_vars(
        self,
        service_header: str,
        header: Optional[str] = PolyaxonServiceHeaders.SERVICE,
        include_secret_key: bool = False,
        include_internal_token: bool = False,
        include_agent_token: bool = False,
        authentication_type: Optional[str] = None,
        external_host: bool = False,
        log_level: Optional[str] = None,
        polyaxon_default_secret_ref: Optional[str] = None,
        polyaxon_agent_secret_ref: Optional[str] = None,
        api_version: Optional[str] = None,
        use_proxy_env_vars_use_in_ops: Optional[bool] = None,
    ) -> List[k8s_schemas.V1EnvVar]:
        api_host = self.get_api_host(external_host)
        polyaxon_default_secret_ref = (
            polyaxon_default_secret_ref or settings.AGENT_CONFIG.app_secret_name
        )
        polyaxon_agent_secret_ref = (
            polyaxon_agent_secret_ref or settings.AGENT_CONFIG.agent_secret_name
        )
        use_proxy_env_vars_use_in_ops = (
            use_proxy_env_vars_use_in_ops
            if use_proxy_env_vars_use_in_ops is not None
            else settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops
        )
        api_version = api_version or VERSION_V1
        env_vars = self._get_base_env_vars(
            namespace=self.namespace,
            resource_name=self.resource_name,
            use_proxy_env_vars_use_in_ops=use_proxy_env_vars_use_in_ops,
        ) + [
            self._get_env_var(name=EV_KEYS_HOST, value=api_host),
            self._get_env_var(name=EV_KEYS_IS_MANAGED, value=True),
            self._get_env_var(name=EV_KEYS_API_VERSION, value=api_version),
            self._get_run_instance_env_var(self.run_instance),
        ]
        if log_level:
            env_vars.append(self._get_env_var(name=EV_KEYS_LOG_LEVEL, value=log_level))
        if header:
            env_vars.append(
                self._get_env_var(
                    name=EV_KEYS_HEADER,
                    value=PolyaxonServiceHeaders.get_header(header),
                )
            )
        if service_header:
            env_vars.append(
                self._get_env_var(
                    name=EV_KEYS_HEADER_SERVICE, value=get_enum_value(service_header)
                )
            )
        if include_secret_key:
            env_vars.append(
                self._get_from_secret(
                    key_name=EV_KEYS_SECRET_KEY,
                    secret_key_name=EV_KEYS_SECRET_KEY,
                    secret_ref_name=polyaxon_default_secret_ref,
                )
            )
        internal = False
        if include_internal_token and polyaxon_default_secret_ref:
            internal = True
            env_vars.append(
                self._get_from_secret(
                    EV_KEYS_SECRET_INTERNAL_TOKEN,
                    EV_KEYS_SECRET_INTERNAL_TOKEN,
                    secret_ref_name=polyaxon_default_secret_ref,
                )
            )
        if include_agent_token and polyaxon_agent_secret_ref:
            if internal:
                raise PolyaxonConverterError(
                    "A service cannot have internal token and agent token."
                )
            env_vars.append(
                self._get_from_secret(
                    EV_KEYS_AUTH_TOKEN,
                    EV_KEYS_AUTH_TOKEN,
                    secret_ref_name=polyaxon_agent_secret_ref,
                )
            )
        if authentication_type:
            env_vars.append(
                self._get_env_var(
                    name=EV_KEYS_AUTHENTICATION_TYPE, value=authentication_type
                )
            )
        return env_vars



import os

from typing import Optional

from clipped.formatting import Printer
from clipped.utils.bools import to_bool

from polyaxon import settings
from polyaxon.env_vars.getters.project import get_project_or_local
from polyaxon.env_vars.keys import (
    EV_KEYS_COLLECT_ARTIFACTS,
    EV_KEYS_COLLECT_RESOURCES,
    EV_KEYS_RUN_INSTANCE,
)
from polyaxon.exceptions import PolyaxonClientException
from polyaxon.managers.run import RunConfigManager


def get_run_or_local(run_uuid=None, is_cli: bool = False):
    if run_uuid:
        return run_uuid
    if is_cli:
        return RunConfigManager.get_config_or_raise().uuid

    try:
        run = RunConfigManager.get_config()
    except TypeError:
        Printer.error(
            "Found an invalid run config or run config cache, "
            "if you are using Polyaxon CLI please run: "
            "`polyaxon config purge --cache-only`",
            sys_exit=True,
        )
    if run:
        return run.uuid
    return None


def get_project_run_or_local(project=None, run_uuid=None, is_cli: bool = True):
    user, project_name = get_project_or_local(project, is_cli=is_cli)
    run_uuid = get_run_or_local(run_uuid, is_cli=is_cli)
    return user, project_name, run_uuid


def get_collect_artifacts(arg: Optional[bool] = None, default: Optional[bool] = None):
    """If set, Polyaxon will collect artifacts"""
    return (
        arg
        if arg is not None
        else to_bool(os.getenv(EV_KEYS_COLLECT_ARTIFACTS, default), handle_none=True)
    )


def get_collect_resources(arg: Optional[bool] = None, default: Optional[bool] = None):
    """If set, Polyaxon will collect resources"""
    return (
        arg
        if arg is not None
        else to_bool(os.getenv(EV_KEYS_COLLECT_RESOURCES, default), handle_none=True)
    )


def get_log_level():
    """If set on the polyaxonfile it will return the log level."""
    return settings.CLIENT_CONFIG.log_level


def get_run_info(run_instance: Optional[str] = None):
    run_instance = run_instance or os.getenv(EV_KEYS_RUN_INSTANCE, None)
    if not run_instance:
        raise PolyaxonClientException(
            "Could not get run info, "
            "please make sure this is run is correctly started by Polyaxon."
        )

    parts = run_instance.split(".")
    if not len(parts) == 4:
        raise PolyaxonClientException(
            "run instance is invalid `{}`, "
            "please make sure this is run is correctly started by Polyaxon.".format(
                run_instance
            )
        )
    return parts[0], parts[1], parts[-1]



from typing import Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.lifecycle import V1StatusCondition
from polyaxon.schemas.base import BaseResponseModel


class V1AgentStatusBodyRequest(BaseResponseModel):
    owner: Optional[StrictStr]
    uuid: Optional[UUIDStr]
    condition: Optional[V1StatusCondition]



from clipped.utils.enums import PEnum


class DeploymentTypes(str, PEnum):
    KUBERNETES = "kubernetes"
    MINIKUBE = "minikube"
    MICRO_K8S = "microk8s"
    DOCKER_COMPOSE = "docker-compose"
    DOCKER = "docker"
    HEROKU = "heroku"


class DeploymentCharts(str, PEnum):
    PLATFORM = "platform"
    AGENT = "agent"
    GATEWAY = "gateway"



from typing import Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr, constr

from polyaxon.schemas.base import NAME_REGEX, BaseSchemaModel


class RootUserConfig(BaseSchemaModel):
    username: Optional[constr(regex=NAME_REGEX)]  # type: ignore[valid-type]
    password: Optional[StrictStr]
    email: Optional[EmailStr]



from polyaxon.sdk.async_client.api_client import AsyncApiClient
from polyaxon.sdk.sync_client.api_client import ApiClient


class BaseApi:
    def __init__(self, api_client=None, is_async=False):
        if api_client is None:
            api_client = (
                AsyncApiClient.get_default() if is_async else ApiClient.get_default()
            )
        self.api_client = api_client



from typing import List, Union

import click

from clipped.formatting import Printer
from clipped.utils.lists import to_list

from polyaxon.connections import V1Connection, V1ConnectionKind


@click.group()
def clean_artifacts():
    pass


def _delete(
    subpath: Union[str, List[str]],
    connection_name: str,
    connection_kind: str,
    is_file: bool,
):
    from polyaxon.fs.fs import get_fs_from_name
    from polyaxon.fs.manager import delete_file_or_dir

    subpath = to_list(subpath, check_none=True)
    fs = get_fs_from_name(connection_name=connection_name)
    for sp in subpath:
        delete_file_or_dir(
            fs=fs,
            subpath=sp,
            is_file=is_file,
        )
    Printer.success(
        "{} subpath was cleaned, subpath: `{}`".format(connection_kind, subpath)
    )


@clean_artifacts.command()
@click.option("--connection-name", help="The connection name.")
@click.option("-sp", "--subpath", multiple=True, help="The s3 subpath to clean.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
def s3(connection_name, subpath, is_file):
    """Delete an s3 subpath."""
    _delete(
        subpath=subpath,
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.S3,
        is_file=is_file,
    )


@clean_artifacts.command()
@click.option("--connection-name", help="The connection name.")
@click.option("-sp", "--subpath", multiple=True, help="The gcs subpath to clean.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
def gcs(connection_name, subpath, is_file):
    """Delete a gcs subpath."""
    _delete(
        subpath=subpath,
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.GCS,
        is_file=is_file,
    )


@clean_artifacts.command()
@click.option("--connection-name", help="The connection name.")
@click.option("-sp", "--subpath", multiple=True, help="The wasb subpath to clean.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
def wasb(connection_name, subpath, is_file):
    """Delete a wasb path context."""
    _delete(
        subpath=subpath,
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.WASB,
        is_file=is_file,
    )


@clean_artifacts.command()
@click.option("--connection-name", help="The connection name.")
@click.option("-sp", "--subpath", multiple=True, help="The volume subpath to clean.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
def volume_claim(connection_name, subpath, is_file):
    """Delete a volume path context."""
    _delete(
        subpath=subpath,
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.VOLUME_CLAIM,
        is_file=is_file,
    )


@clean_artifacts.command()
@click.option("--connection-name", help="The connection name.")
@click.option("-sp", "--subpath", multiple=True, help="The host subpath to clean.")
@click.option(
    "--is-file",
    is_flag=True,
    default=False,
    help="whether or not to use the basename of the key.",
)
def host_path(connection_name, subpath, is_file):
    """Delete a host path context."""
    _delete(
        subpath=subpath,
        connection_name=connection_name,
        connection_kind=V1ConnectionKind.HOST_PATH,
        is_file=is_file,
    )



import datetime

from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.connections import V1ConnectionKind
from polyaxon.schemas.base import BaseResponseModel


class V1ConnectionResponse(BaseResponseModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    agent: Optional[StrictStr]
    description: Optional[StrictStr]
    tags: Optional[List[StrictStr]]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    live_state: Optional[int]
    kind: Optional[V1ConnectionKind]



from typing import Optional
from typing_extensions import Literal

from clipped.types.ref_or_obj import BoolOrRef, DatetimeOrRef, IntOrRef
from pydantic import Field, StrictStr

from polyaxon.polyflow.schedules.kinds import V1ScheduleKind
from polyaxon.schemas.base import BaseSchemaModel


class V1CronSchedule(BaseSchemaModel):
    """Cron schedules is an interface to trigger components repeatedly using a cron definition.

    Args:
        kind: str, should be equal to `cron`
        start_at: datetime, optional
        end_at: datetime, optional
        max_runs: int, optional
        cron: str, required
        depends_on_past: bool, optional


    ## YAML usage

    ```yaml
    >>> schedule:
    >>>   kind:
    >>>   startAt:
    >>>   endAt:
    >>>   maxRuns:
    >>>   cron:
    >>>   dependsOnPast:
    ```

    ## Python usage

    ```python
    >>> from datetime import datetime
    >>> from polyaxon.polyflow import V1CronSchedule
    >>> schedule = V1CronSchedule(
    >>>   start_at=datetime(...),
    >>>   end_at=datetime(...),
    >>>   max_runs=20,
    >>>   cron="0 0 * * *",
    >>>   dependsOnPast=False,
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this schedule is a cron schedule.

    If you are using the python client to create the schedule,
    this field is not required and is set by default.

    ```yaml
    >>> run:
    >>>   kind: cron
    ```

    ### startAt

    Optional field to set the start time for kicking the first execution,
    all following executions will be relative to this time.

    ```yaml
    >>> run:
    >>>   startAt: "2019-06-24T21:20:07+00:00"
    ```

    ### endAt

    Optional field to set the end time for stopping this schedule.

    ```yaml
    >>> run:
    >>>   endAt: "2019-06-24T21:20:07+00:00"
    ```

    ### maxRuns

    The maximum number of times to execute the component.
    If used with end date, the schedule will terminate if one of the conditions is met.

    ```yaml
    >>> run:
    >>>   maxRuns: 10
    ```

    ### cron

    The cron expression to generate the component executions.

    ```yaml
    >>> run:
    >>>   cron: "0 0 * * *"
    ```

    ### dependsOnPast

    Optional field to tell Polyaxon to check if the
    previous execution was done before scheduling a new one, by default this is set to `false`.

    ```yaml
    >>> run:
    >>>   dependsOnPast: true
    ```
    """

    _IDENTIFIER = V1ScheduleKind.CRON
    _USE_DISCRIMINATOR = True

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    cron: StrictStr
    start_at: Optional[DatetimeOrRef] = Field(alias="startAt")
    end_at: Optional[DatetimeOrRef] = Field(alias="endAt")
    max_runs: Optional[IntOrRef] = Field(alias="maxRuns")
    depends_on_past: Optional[BoolOrRef] = Field(alias="dependsOnPast")



from polyaxon.deploy.operators.cmd_operator import CmdOperator


class PipOperator(CmdOperator):
    CMD = "pip"

    @classmethod
    def params(cls, args):
        params = [cls.CMD] + args
        return params

    @classmethod
    def check(cls):
        command_exist = cls.execute(args=["-V"])
        if not command_exist:
            return False
        return True

    @classmethod
    def execute(cls, args, stream=False):
        params = cls.params(args)
        return cls._execute(params=params, env=None, stream=stream)



import click

OPTIONS_OWNER = {
    "args": ["--owner", "-o"],
    "kwargs": dict(
        type=str,
        help="Name of the owner/namespace, "
        "if not provided it will default to the namespace of the current user.",
    ),
}

OPTIONS_NAME = {
    "args": ["--name", "-n"],
}

OPTIONS_PROJECT = {
    "args": ["--project", "-p"],
    "kwargs": dict(type=str, help="The project name, e.g. 'mnist' or 'acme/mnist'."),
}

OPTIONS_MODEL_VERSION = {
    "args": ["--version", "-ver"],
    "kwargs": dict(
        type=str,
        help="The component version, e.g. 'v1.3.4' or 'rc1' or 'latest'.",
    ),
}

OPTIONS_COMPONENT_VERSION = {
    "args": ["--version", "-ver"],
    "kwargs": dict(
        type=str,
        help="The component version, e.g. 'v1.3.4' or 'rc1' or 'latest'.",
    ),
}


OPTIONS_ARTIFACT_VERSION = {
    "args": ["--version", "-ver"],
    "kwargs": dict(
        type=str,
        help="The artifact version, e.g. 'v1.3.4' or 'rc1' or 'latest'.",
    ),
}

OPTIONS_RUN_UID = {
    "args": ["--uid", "-uid"],
    "kwargs": dict(type=str, help="The run uuid."),
}


OPTIONS_RUN_OFFLINE = {
    "args": ["--offline"],
    "kwargs": dict(
        is_flag=True,
        default=False,
        help="To use the offline store, for resolving runs, if it exists.",
    ),
}

OPTIONS_RUN_OFFLINE_PATH_FROM = {
    "args": [
        "--path",
        "--path-from",
    ],
    "kwargs": dict(
        type=click.Path(exists=False),
        help="Optional offline root path to use where runs are persisted, "
        "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
    ),
}
OPTIONS_RUN_OFFLINE_PATH_TO = {
    "args": [
        "--path",
        "--path-to",
    ],
    "kwargs": dict(
        type=click.Path(exists=False),
        help="Optional offline root path to use where runs are persisted, "
        "default value is taken from the env var: `POLYAXON_OFFLINE_ROOT`.",
    ),
}



import os

from datetime import datetime
from requests import HTTPError
from typing import Dict, List, Optional, Tuple, Union

from clipped.utils.json import orjson_dumps
from clipped.utils.paths import check_or_create_path, delete_path
from clipped.utils.query_params import get_query_params
from clipped.utils.tz import now
from clipped.utils.validation import validate_tags

from polyaxon.client.client import PolyaxonClient
from polyaxon.client.decorators import client_handler, get_global_or_inline_config
from polyaxon.constants.globals import DEFAULT
from polyaxon.contexts import paths as ctx_paths
from polyaxon.env_vars.getters.user import get_local_owner
from polyaxon.exceptions import PolyaxonClientException
from polyaxon.lifecycle import V1ProjectVersionKind, V1StageCondition, V1Stages
from polyaxon.logger import logger
from polyaxon.schemas.responses.v1_list_project_versions_response import (
    V1ListProjectVersionsResponse,
)
from polyaxon.schemas.responses.v1_project import V1Project
from polyaxon.schemas.responses.v1_project_version import V1ProjectVersion
from polyaxon.sdk.exceptions import ApiException
from polyaxon.utils.fqn_utils import get_entity_full_name, get_entity_info
from traceml.artifacts import V1RunArtifact


class ProjectClient:
    """ProjectClient is a client to communicate with Polyaxon projects endpoints.

    If no values are passed to this class,
    Polyaxon will try to resolve the owner and project from the environment:
     * If you have a configured CLI, Polyaxon will use the configuration of the cli.
     * If you have a cached project using the CLI,
       the client will default to that cached project unless you override the values.
     * If you use this client in the context of a job or a service managed by Polyaxon,
       a configuration will be available to resolve the values based on that run.

    If you intend to create a new project instance or to list projects,
    only the `owner` parameter is required.

    Properties:
        project: str.
        owner: str.
        project_data: V1Project.

    Args:
        owner: str, optional, the owner is the username or
             the organization name owning this project.
        project: str, optional, project name.
        client: [PolyaxonClient](/docs/core/python-library/polyaxon-client/), optional,
             an instance of a configured client, if not passed,
             a new instance will be created based on the available environment.
        is_offline: bool, optional,
             To trigger the offline mode manually instead of depending on `POLYAXON_IS_OFFLINE`.
        no_op: bool, optional,
             To set the NO_OP mode manually instead of depending on `POLYAXON_NO_OP`.

    Raises:
        PolyaxonClientException: If no owner is passed and Polyaxon cannot
            resolve an owner from the environment.
    """

    @client_handler(check_no_op=True)
    def __init__(
        self,
        owner: Optional[str] = None,
        project: Optional[str] = None,
        client: Optional[PolyaxonClient] = None,
        is_offline: Optional[bool] = None,
        no_op: Optional[bool] = None,
        manual_exceptions_handling: bool = False,
    ):
        self._manual_exceptions_handling = manual_exceptions_handling
        self._is_offline = get_global_or_inline_config(
            config_key="is_offline", config_value=is_offline, client=client
        )
        self._no_op = get_global_or_inline_config(
            config_key="no_op", config_value=no_op, client=client
        )

        if self._no_op:
            return

        if not owner and project:
            owner, project = get_entity_info(
                get_entity_full_name(owner=owner, entity=project)
            )

        if not owner:
            owner = get_local_owner()
        if not owner:
            raise PolyaxonClientException("Please provide a valid owner.")

        self._client = client
        self._owner = owner or DEFAULT
        self._project = project
        self._project_data = V1Project.construct()

    @property
    def client(self):
        if self._client:
            return self._client
        self._client = PolyaxonClient()
        return self._client

    @property
    def owner(self):
        return self._owner

    @property
    def project(self):
        return self._project

    @property
    def project_data(self):
        return self._project_data

    @client_handler(check_no_op=True, check_offline=True)
    def refresh_data(self):
        """Fetches the project data from the api."""
        self._project_data = self.client.projects_v1.get_project(
            self.owner, self.project
        )
        if self._project_data.owner is None:
            self._project_data.owner = self.owner

    @client_handler(check_no_op=True, check_offline=True)
    def create(self, data: Union[Dict, V1Project]) -> V1Project:
        """Creates a new project based on the data passed.

        [Project API](/docs/api/#operation/CreateProject)

        Args:
            data: dict or V1Project, required.

        Returns:
            V1Project, project instance from the response.
        """
        self._project_data = self.client.projects_v1.create_project(
            self.owner,
            data,
            async_req=False,
        )
        self._project_data.owner = self.owner
        self._project = self._project_data.name
        return self._project_data

    @client_handler(check_no_op=True, check_offline=True)
    def get_or_create(self):
        try:
            self.refresh_data()
        except ApiException:
            self.project_data.name = self.project
            self.create(self.project_data)

    @client_handler(check_no_op=True, check_offline=True)
    def list(
        self,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[V1Project]:
        """Lists projects under the current owner.

        [Project API](/docs/api/#operation/ListProjects)

        Args:
            query: str, optional, query filters, please refer to
                    [Project PQL](/docs/core/query-language/projects/#query)
            sort: str, optional, fields to order by, please refer to
                    [Project PQL](/docs/core/query-language/projects/#sort)
            limit: int, optional, limit of projects to return.
            offset: int, optional, offset pages to paginate projects.

        Returns:
            List[V1Project], list of project instances.
        """
        params = get_query_params(limit=limit, offset=offset, query=query, sort=sort)
        return self.client.projects_v1.list_projects(self.owner, **params)

    @client_handler(check_no_op=True, check_offline=True)
    def delete(self):
        """Deletes project based on the current owner and project."""
        return self.client.projects_v1.delete_project(self.owner, self.project)

    @client_handler(check_no_op=True, check_offline=True)
    def update(self, data: Union[Dict, V1Project]) -> V1Project:
        """Updates a project based on the data passed.

        [Project API](/docs/api/#operation/PatchProject)

        Args:
            data: Dict or V1Project, required.

        Returns:
            V1Project, project instance from the response.
        """
        self._project_data = self.client.projects_v1.patch_project(
            self.owner,
            self.project,
            body=data,
            async_req=False,
        )
        self._project = self._project_data.name
        return self._project_data

    @client_handler(check_no_op=True, check_offline=True)
    def list_runs(
        self,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ):
        """Lists runs under the current owner/project.

        [Run API](/docs/api/#operation/ListRuns)

        Args:
            query: str, optional, query filters, please refer to
                 [Run PQL](/docs/core/query-language/runs/#query)
            sort: str, optional, fields to order by, please refer to
                 [Run PQL](/docs/core/query-language/runs/#sort)
            limit: int, optional, limit of runs to return.
            offset: int, optional, offset pages to paginate runs.

        Returns:
            List[V1Run], list of run instances.
        """
        params = get_query_params(
            limit=limit or 20, offset=offset, query=query, sort=sort
        )
        return self.client.runs_v1.list_runs(self.owner, self.project, **params)

    def _validate_kind(self, kind: V1ProjectVersionKind):
        if kind not in V1ProjectVersionKind:
            raise ValueError(
                "The kind `{}` is not supported, it must be one of the values `{}`".format(
                    kind, V1ProjectVersionKind.to_list()
                )
            )

    @client_handler(check_no_op=True, check_offline=True)
    def list_versions(
        self,
        kind: V1ProjectVersionKind,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> V1ListProjectVersionsResponse:
        """Lists project versions under the current owner/project based on version kind.

        This is a generic function that maps to list:
            * component versions
            * model versions
            * artifact versions

        [Project API](/docs/api/#operation/ListProjectVersions)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            query: str, optional, query filters, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#query)
            sort: str, optional, fields to order by, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#sort)
            limit: int, optional, limit of project versions to return.
            offset: int, optional, offset pages to paginate project versions.

        Returns:
            List[V1ProjectVersion], list of versions.
        """
        self._validate_kind(kind)
        params = get_query_params(
            limit=limit or 20, offset=offset, query=query, sort=sort
        )
        return self.client.projects_v1.list_versions(
            self.owner, self.project, kind, **params
        )

    @client_handler(check_no_op=True, check_offline=True)
    def list_component_versions(
        self,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> V1ListProjectVersionsResponse:
        """Lists component versions under the current owner/project.

        [Project API](/docs/api/#operation/ListProjectVersions)

        Args:
            query: str, optional, query filters, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#query)
            sort: str, optional, fields to order by, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#sort)
            limit: int, optional, limit of project versions to return.
            offset: int, optional, offset pages to paginate project versions.

        Returns:
            List[V1ProjectVersion], list of component versions.
        """
        return self.list_versions(
            kind=V1ProjectVersionKind.COMPONENT,
            query=query,
            sort=sort,
            limit=limit,
            offset=offset,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def list_model_versions(
        self,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> V1ListProjectVersionsResponse:
        """Lists model versions under the current owner/project.

        [Project API](/docs/api/#operation/ListProjectVersions)

        Args:
            query: str, optional, query filters, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#query)
            sort: str, optional, fields to order by, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#sort)
            limit: int, optional, limit of project versions to return.
            offset: int, optional, offset pages to paginate project versions.

        Returns:
            List[V1ProjectVersion], list of model versions.
        """
        return self.list_versions(
            kind=V1ProjectVersionKind.MODEL,
            query=query,
            sort=sort,
            limit=limit,
            offset=offset,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def list_artifact_versions(
        self,
        query: Optional[str] = None,
        sort: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> V1ListProjectVersionsResponse:
        """Lists artifact versions under the current owner/project.

        [Project API](/docs/api/#operation/ListProjectVersions)

        Args:
            query: str, optional, query filters, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#query)
            sort: str, optional, fields to order by, please refer to
                 [Project Version PQL](/docs/core/query-language/project-versions/#sort)
            limit: int, optional, limit of project versions to return.
            offset: int, optional, offset pages to paginate project versions.

        Returns:
            List[V1ProjectVersion], list of artifact versions.
        """
        return self.list_versions(
            kind=V1ProjectVersionKind.ARTIFACT,
            query=query,
            sort=sort,
            limit=limit,
            offset=offset,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def get_version(self, kind: V1ProjectVersionKind, version: str) -> V1ProjectVersion:
        """Gets a project version under the current owner/project based on version kind.

        This is a generic function that maps to get:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/GetVersion)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.

        Returns:
            V1ProjectVersion.
        """
        self._validate_kind(kind)
        response = self.client.projects_v1.get_version(
            self.owner, self.project, kind, version
        )
        if response.kind != kind:
            raise PolyaxonClientException("This version is not of kind `%s`." % kind)
        return response

    @client_handler(check_no_op=True, check_offline=True)
    def get_component_version(self, version: str) -> V1ProjectVersion:
        """Gets a component version under the current owner/project.

        [Project API](/docs/api/#operation/GetVersion)

        Args:
            version: str, required, the version name/tag.

        Returns:
            V1ProjectVersion, component version.
        """
        return self.get_version(kind=V1ProjectVersionKind.COMPONENT, version=version)

    @client_handler(check_no_op=True, check_offline=True)
    def get_model_version(self, version: str) -> V1ProjectVersion:
        """Gets a model version under the current owner/project.

        [Project API](/docs/api/#operation/GetVersion)

        Args:
            version: str, required, the version name/tag.

        Returns:
            V1ProjectVersion, model version.
        """
        return self.get_version(kind=V1ProjectVersionKind.MODEL, version=version)

    @client_handler(check_no_op=True, check_offline=True)
    def get_artifact_version(self, version: str) -> V1ProjectVersion:
        """Gets an artifact version under the current owner/project.

        [Project API](/docs/api/#operation/GetVersion)

        Args:
            version: str, required, the version name/tag.

        Returns:
            V1ProjectVersion, artifact version.
        """
        return self.get_version(kind=V1ProjectVersionKind.ARTIFACT, version=version)

    @client_handler(check_no_op=True, check_offline=True)
    def get_version_stages(
        self, kind: V1ProjectVersionKind, version: str
    ) -> Tuple[str, List[V1StageCondition]]:
        """Gets a project version stages under the current owner/project based on version kind.

        This is a generic function that maps to get:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/GetVersionStages)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.

        Returns:
            Tuple[str, List[V1StageCondition]]
        """
        self._validate_kind(kind)
        response = self.client.projects_v1.get_version_stages(
            self.owner, self.project, kind, version
        )
        return response.stage, response.stage_conditions

    @client_handler(check_no_op=True, check_offline=True)
    def get_component_version_stages(
        self, version: str
    ) -> Tuple[str, List[V1StageCondition]]:
        """Gets a component version stages under the current owner/project.

        [Project API](/docs/api/#operation/GetVersionStages)

        Args:
            version: str, required, the version name/tag.

        Returns:
            Tuple[str, List[V1StageCondition]]
        """
        return self.get_version_stages(
            kind=V1ProjectVersionKind.COMPONENT, version=version
        )

    @client_handler(check_no_op=True, check_offline=True)
    def get_model_version_stages(
        self, version: str
    ) -> Tuple[str, List[V1StageCondition]]:
        """Gets a model version under the current owner/project.

        [Project API](/docs/api/#operation/GetVersionStages)

        Args:
            version: str, required, the version name/tag.

        Returns:
            Tuple[str, List[V1StageCondition]]
        """
        return self.get_version_stages(kind=V1ProjectVersionKind.MODEL, version=version)

    @client_handler(check_no_op=True, check_offline=True)
    def get_artifact_version_stages(
        self, version: str
    ) -> Tuple[str, List[V1StageCondition]]:
        """Gets an artifact version under the current owner/project.

        [Project API](/docs/api/#operation/GetVersionStages)

        Args:
            version: str, required, the version name/tag.

        Returns:
            Tuple[str, List[V1StageCondition]]
        """
        return self.get_version_stages(
            kind=V1ProjectVersionKind.ARTIFACT, version=version
        )

    @client_handler(check_no_op=True, check_offline=True)
    def create_version(
        self,
        kind: V1ProjectVersionKind,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Creates a project version based on the data passed based on version kind.

        This is a generic function based on the kind passed and creates a:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/CreateVersion)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion.
        """
        self._validate_kind(kind)
        if isinstance(data, V1ProjectVersion):
            data.kind = kind
        elif isinstance(data, dict):
            data["kind"] = kind
        return self.client.projects_v1.create_version(
            self.owner,
            self.project,
            kind,
            body=data,
            async_req=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def create_component_version(
        self,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Creates a component version based on the data passed.

        [Project API](/docs/api/#operation/CreateVersion)

        Args:
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, component version.
        """
        return self.create_version(
            kind=V1ProjectVersionKind.COMPONENT,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def create_model_version(
        self,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Creates a model version based on the data passed.

        [Project API](/docs/api/#operation/CreateVersion)

        Args:
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, model version.
        """
        return self.create_version(
            kind=V1ProjectVersionKind.MODEL,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def create_artifact_version(
        self,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Creates an artifact version based on the data passed.

        [Project API](/docs/api/#operation/CreateVersion)

        Args:
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, artifact version.
        """
        return self.create_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def patch_version(
        self,
        kind: V1ProjectVersionKind,
        version: str,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Updates a project version based on the data passed and version kind.

        This is a generic function based on the kind passed and patches a:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/PatchVersion)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion.
        """
        self._validate_kind(kind)
        return self.client.projects_v1.patch_version(
            self.owner,
            self.project,
            kind,
            version,
            body=data,
            async_req=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def patch_component_version(
        self,
        version: str,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Updates a component version based on the data passed.

        [Project API](/docs/api/#operation/PatchVersion)

        Args:
            version: str, required, the version name/tag.
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, component version.
        """
        return self.patch_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def patch_model_version(
        self,
        version: str,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Updates a model version based on the data passed.

        [Project API](/docs/api/#operation/PatchVersion)

        Args:
            version: str, required, the version name/tag.
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, model version.
        """
        return self.patch_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def patch_artifact_version(
        self,
        version: str,
        data: Union[Dict, V1ProjectVersion],
    ) -> V1ProjectVersion:
        """Updates an artifact version based on the data passed.

        [Project API](/docs/api/#operation/PatchVersion)

        Args:
            version: str, required, the version name/tag.
            data: Dict or V1ProjectVersion, required.

        Returns:
            V1ProjectVersion, artifact version.
        """
        return self.patch_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            data=data,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def register_version(
        self,
        kind: V1ProjectVersionKind,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        run: Optional[str] = None,
        connection: Optional[str] = None,
        artifacts: Optional[List[str]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Creates or Updates a project version based on the data passed.

        This is a generic function based on the kind passed and registers a:
            * component version
            * model version
            * artifact version

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, optional, the version name/tag.
            description: str, optional, the version description.
            tags: str or List[str], optional.
            content: str or dict, optional, content/metadata (JSON object) of the version.
            run: str, optional, a uuid reference to the run.
            connection: str, optional, a uuid reference to a connection.
            artifacts: List[str], optional, list of artifacts to highlight(requires passing a run)
            force: bool, optional, to force push, i.e. update if exists.

        Returns:
            V1ProjectVersion.
        """
        try:
            self.get_version(kind, version)
            if not force:
                raise PolyaxonClientException(
                    "A {} version {} already exists. "
                    "Please pass the `force` argument or `--force` flag for CLI) "
                    "if you want to push force this version.".format(kind, version)
                )
            to_update = True
        except (ApiException, HTTPError, AttributeError):
            to_update = False

        def _get_content() -> str:
            return content if isinstance(content, str) else orjson_dumps(content)

        if content:
            content = _get_content()
        if tags is not None:
            tags = validate_tags(tags, validate_yaml=True)
        if artifacts is not None:
            artifacts = validate_tags(artifacts, validate_yaml=True)

        if to_update:
            version_config = V1ProjectVersion.construct()
            if description is not None:
                version_config.description = description
            if tags:
                version_config.tags = tags
            if content:
                version_config.content = content  # type: ignore
            if run:
                version_config.run = run
            if artifacts is not None:
                version_config.artifacts = artifacts
            if connection is not None:
                version_config.connection = connection
            return self.patch_version(
                kind=kind,
                version=version,
                data=version_config,
            )
        else:
            version_config = V1ProjectVersion.construct(
                name=version,
                description=description,
                tags=tags,
                run=run,
                artifacts=artifacts,
                connection=connection,
                content=content,
            )
            return self.create_version(kind=kind, data=version_config)

    @client_handler(check_no_op=True, check_offline=True)
    def register_component_version(
        self,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        run: Optional[str] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Creates or Updates a component version based on the data passed.

        Args:
            version: str, optional, the version name/tag.
            description: str, optional, the version description.
            tags: str or List[str], optional.
            content: str or dict, optional, content/metadata (JSON object) of the version.
            run: str, optional, a uuid reference to the run.
            force: bool, optional, to force push, i.e. update if exists.

        Returns:
            V1ProjectVersion, component verison.
        """
        return self.register_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            description=description,
            tags=tags,
            content=content,
            run=run,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def register_model_version(
        self,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        run: Optional[str] = None,
        connection: Optional[str] = None,
        artifacts: Optional[List[str]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Create or Update a model version based on the data passed.

        Args:
            version: str, optional, the version name/tag.
            description: str, optional, the version description.
            tags: str or List[str], optional.
            content: str or dict, optional, content/metadata (JSON object) of the version.
            run: str, optional, a uuid reference to the run.
            connection: str, optional, a uuid reference to a connection.
            artifacts: List[str], optional, list of artifacts to highlight(requires passing a run)
            force: bool, optional, to force push, i.e. update if exists.

        Returns:
            V1ProjectVersion, model version.
        """
        return self.register_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            description=description,
            tags=tags,
            content=content,
            run=run,
            connection=connection,
            artifacts=artifacts,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def register_artifact_version(
        self,
        version: str,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        run: Optional[str] = None,
        connection: Optional[str] = None,
        artifacts: Optional[List[str]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Create or Update an artifact version based on the data passed.

        Args:
            version: str, optional, the version name/tag.
            description: str, optional, the version description.
            tags: str or List[str], optional.
            content: str or dict, optional, content/metadata (JSON object) of the version.
            run: str, optional, a uuid reference to the run.
            connection: str, optional, a uuid reference to a connection.
            artifacts: List[str], optional, list of artifacts to highlight(requires passing a run)
            force: bool, optional, to force push, i.e. update if exists.

        Returns:
            V1ProjectVersion, artifact version.
        """
        return self.register_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            description=description,
            tags=tags,
            content=content,
            run=run,
            connection=connection,
            artifacts=artifacts,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def delete_version(self, kind: V1ProjectVersionKind, version: str):
        """Deletes a project version under the current owner/project.

        This is a generic function based on the kind passed and deletes a:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/DeleteVersion)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
        """
        self._validate_kind(kind)
        logger.info("Deleting {} version: `{}`".format(kind, version))
        return self.client.projects_v1.delete_version(
            self.owner,
            self.project,
            kind,
            version,
            async_req=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def delete_component_version(self, version: str):
        """Deletes a component version under the current owner/project.

        [Project API](/docs/api/#operation/DeleteVersion)

        Args:
            version: str, required, the version name/tag.
        """
        return self.delete_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def delete_model_version(self, version: str):
        """Deletes a model version under the current owner/project.

        [Project API](/docs/api/#operation/DeleteVersion)

        Args:
            version: str, required, the version name/tag.
        """
        return self.delete_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def delete_artifact_version(self, version: str):
        """Deletes an artifact version under the current owner/project.

        [Project API](/docs/api/#operation/DeleteVersion)

        Args:
            version: str, required, the version name/tag.
        """
        return self.delete_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def stage_version(
        self,
        kind: V1ProjectVersionKind,
        version: str,
        stage: Union[str, V1Stages],
        reason: Optional[str] = None,
        message: Optional[str] = None,
        last_transition_time: Optional[datetime] = None,
        last_update_time: Optional[datetime] = None,
    ):
        """Creates a new a project version stage.


        This is a generic function based on the kind passed and stages a:
            * component version
            * model version
            * artifact version

        [Project API](/docs/api/#operation/CreateVersionStage)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
            stage: str, a valid [Stages](/docs/core/specification/lifecycle/) value.
            reason: str, optional, reason or service issuing the stage change.
            message: str, optional, message to log with this status.
            last_transition_time: datetime, default `now`.
            last_update_time: datetime, default `now`.
        """
        self._validate_kind(kind)
        current_date = now()
        stage_condition = V1StageCondition.construct(
            type=stage,
            status=True,
            reason=reason or "ClientStageUpdate",
            message=message,
            last_transition_time=last_transition_time or current_date,
            last_update_time=last_update_time or current_date,
        )
        return self.client.projects_v1.create_version_stage(
            self.owner,
            self.project,
            kind,
            version,
            body={"condition": stage_condition},
            async_req=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def stage_component_version(
        self,
        version: str,
        stage: Union[str, V1Stages],
        reason: Optional[str] = None,
        message: Optional[str] = None,
        last_transition_time: Optional[datetime] = None,
        last_update_time: Optional[datetime] = None,
    ):
        """Creates a new a component version stage.

        [Project API](/docs/api/#operation/CreateVersionStage)

        Args:
            version: str, required, the version name/tag.
            stage: str, a valid [Stages](/docs/core/specification/lifecycle/) value.
            reason: str, optional, reason or service issuing the status change.
            message: str, optional, message to log with this status.
            last_transition_time: datetime, default `now`.
            last_update_time: datetime, default `now`.
        """
        return self.stage_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            stage=stage,
            reason=reason,
            message=message,
            last_transition_time=last_transition_time,
            last_update_time=last_update_time,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def stage_model_version(
        self,
        version: str,
        stage: Union[str, V1Stages],
        reason: Optional[str] = None,
        message: Optional[str] = None,
        last_transition_time: Optional[datetime] = None,
        last_update_time: Optional[datetime] = None,
    ):
        """Creates a new a model version stage.

        [Project API](/docs/api/#operation/CreateVersionStage)

        Args:
            version: str, required, the version name/tag.
            stage: str, a valid [Stages](/docs/core/specification/lifecycle/) value.
            reason: str, optional, reason or service issuing the status change.
            message: str, optional, message to log with this status.
            last_transition_time: datetime, default `now`.
            last_update_time: datetime, default `now`.
        """
        return self.stage_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            stage=stage,
            reason=reason,
            message=message,
            last_transition_time=last_transition_time,
            last_update_time=last_update_time,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def stage_artifact_version(
        self,
        version: str,
        stage: Union[str, V1Stages],
        reason: Optional[str] = None,
        message: Optional[str] = None,
        last_transition_time: Optional[datetime] = None,
        last_update_time: Optional[datetime] = None,
    ):
        """Creates a new an artifact version stage.

        [Project API](/docs/api/#operation/CreateVersionStage)

        Args:
            version: str, required, the version name/tag.
            stage: str, a valid [Stages](/docs/core/specification/lifecycle/) value.
            reason: str, optional, reason or service issuing the status change.
            message: str, optional, message to log with this status.
            last_transition_time: datetime, default `now`.
            last_update_time: datetime, default `now`.
        """
        return self.stage_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            stage=stage,
            reason=reason,
            message=message,
            last_transition_time=last_transition_time,
            last_update_time=last_update_time,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def transfer_version(
        self, kind: V1ProjectVersionKind, version: str, to_project: str
    ):
        """Transfers the version to a project under the same owner/organization.

        This is a generic function based on the kind passed and transfers a:
            * component version
            * model version
            * artifact version

        [Run API](/docs/api/#operation/TransferVersion)

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
            to_project: str, required, the destination project to transfer the version to.
        """

        self._validate_kind(kind)
        return self.client.projects_v1.transfer_version(
            self.owner,
            self.project,
            kind,
            version,
            body={"project": to_project},
            async_req=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def transfer_component_version(self, version: str, to_project: str):
        """Transfers the component version to a project under the same owner/organization.

        [Run API](/docs/api/#operation/TransferVersion)

        Args:
            version: str, required, the version name/tag.
            to_project: str, required, the destination project to transfer the version to.
        """
        return self.transfer_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            to_project=to_project,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def transfer_model_version(self, version: str, to_project: str):
        """Transfers the model version to a project under the same owner/organization.

        [Run API](/docs/api/#operation/TransferVersion)

        Args:
            version: str, required, the version name/tag.
            to_project: str, required, the destination project to transfer the version to.
        """
        return self.transfer_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            to_project=to_project,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def transfer_artifact_version(self, version: str, to_project: str):
        """Transfers the artifact version to a project under the same owner/organization.

        [Run API](/docs/api/#operation/TransferVersion)

        Args:
            version: str, required, the version name/tag.
            to_project: str, required, the destination project to transfer the version to.
        """
        return self.transfer_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            to_project=to_project,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def copy_version(
        self,
        kind: V1ProjectVersionKind,
        version: str,
        to_project: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Copies the version to the same project or to a destination project.

        If `to_project` is provided,
        the version will be copied to the destination project under the same owner/organization.

        If `name` is provided the version will be copied with the new name,
        otherwise the copied version will be have a suffix `-copy`.

        This is a generic function based on the kind passed and copies a:
            * component version
            * model version
            * artifact version

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
            to_project: str, optional, the destination project to copy the version to.
            name: str, optional, the name to use for registering the copied version,
                 default value is the original version's name with `-copy` prefix.
            description: str, optional, the version description,
                 default value is the original version's description.
            tags: str or List[str], optional, the version description,
                 default value is the original version's description.
            content: str or dict, optional, content/metadata (JSON object) of the version,
                 default value is the original version's content.
            force: bool, optional, to force push, i.e. update if exists.
        """
        original_version = self.get_version(kind, version)
        version = name if name else "{}-copy".format(version)
        return ProjectClient(
            owner=self.owner,
            project=to_project or self.project,
            client=self.client,
        ).register_version(
            kind=kind,
            version=version,
            description=description or original_version.description,
            tags=tags or original_version.tags,
            content=content or original_version.content,
            run=original_version.run,
            connection=original_version.connection,
            artifacts=original_version.artifacts,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def copy_component_version(
        self,
        version: str,
        to_project: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Copies the component version to the same project or to a destination project.

        If `to_project` is provided,
        the version will be copied to the destination project under the same owner/organization.

        If `name` is provided the version will be copied with the new name,
        otherwise the copied version will be have a suffix `-copy`.

        Args:
            version: str, required, the version name/tag.
            to_project: str, optional, the destination project to copy the version to.
            name: str, optional, the name to use for registering the copied version,
                 default value is the original version's name with `-copy` prefix.
            description: str, optional, the version description,
                 default value is the original version's description.
            tags: str or List[str], optional, the version description,
                 default value is the original version's description.
            content: str or dict, optional, content/metadata (JSON object) of the version,
                 default value is the original version's content.
            force: bool, optional, to force push, i.e. update if exists.
        """
        return self.copy_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            to_project=to_project,
            name=name,
            description=description,
            tags=tags,
            content=content,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def copy_model_version(
        self,
        version: str,
        to_project: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Copies the model version to the same project or to a destination project.

        If `to_project` is provided,
        the version will be copied to the destination project under the same owner/organization.

        If `name` is provided the version will be copied with the new name,
        otherwise the copied version will be have a suffix `-copy`.

        Args:
            version: str, required, the version name/tag.
            to_project: str, optional, the destination project to copy the version to.
            name: str, optional, the name to use for registering the copied version,
                 default value is the original version's name with `-copy` prefix.
            description: str, optional, the version description,
                 default value is the original version's description.
            tags: str or List[str], optional, the version description,
                 default value is the original version's description.
            content: str or dict, optional, content/metadata (JSON object) of the version,
                 default value is the original version's content.
            force: bool, optional, to force push, i.e. update if exists.
        """
        return self.copy_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            to_project=to_project,
            name=name,
            description=description,
            tags=tags,
            content=content,
            force=force,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def copy_artifact_version(
        self,
        version: str,
        to_project: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
        content: Optional[Union[str, Dict]] = None,
        force: bool = False,
    ) -> V1ProjectVersion:
        """Copies the artifact version to the same project or to a destination project.

        If `to_project` is provided,
        the version will be copied to the destination project under the same owner/organization.

        If `name` is provided the version will be copied with the new name,
        otherwise the copied version will be have a suffix `-copy`.

        Args:
            version: str, required, the version name/tag.
            to_project: str, optional, the destination project to copy the version to.
            name: str, optional, the name to use for registering the copied version,
                 default value is the original version's name with `-copy` prefix.
            description: str, optional, the version description,
                 default value is the original version's description.
            tags: str or List[str], optional, the version description,
                 default value is the original version's description.
            content: str or dict, optional, content/metadata (JSON object) of the version,
                 default value is the original version's content.
            force: bool, optional, to force push, i.e. update if exists.
        """
        return self.copy_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            to_project=to_project,
            name=name,
            description=description,
            tags=tags,
            content=content,
            force=force,
        )

    @client_handler(check_no_op=True)
    def persist_version(self, config: V1ProjectVersion, path: str):
        """Persists a version to a local path.

        Args:
            config: V1ProjectVersion, the version config to persist.
            path: str, the path where to persist the version config.
        """
        if not config:
            logger.debug(
                "Persist offline run call failed. "
                "Make sure that the offline mode is enabled and that run_data is provided."
            )
            return
        if not path or not os.path.exists(path):
            check_or_create_path(path, is_dir=True)
        version_path = "{}/{}".format(path, ctx_paths.CONTEXT_LOCAL_VERSION)
        with open(version_path, "w") as config_file:
            config_file.write(
                orjson_dumps(self.client.sanitize_for_serialization(config))
            )
        if not config.content:
            return
        if config.kind == V1ProjectVersionKind.COMPONENT:
            version_path = "{}/{}".format(path, ctx_paths.CONTEXT_LOCAL_POLYAXONFILE)
        else:
            # Persist content metadata as content.json file
            version_path = "{}/{}".format(path, ctx_paths.CONTEXT_LOCAL_CONTENT)
        with open(version_path, "w") as config_file:
            config_file.write(config.content)

    @client_handler(check_no_op=True, check_offline=True)
    def download_artifacts_for_version(self, config: V1ProjectVersion, path: str):
        """Collects and downloads all artifacts and assets linked to a version.

        Args:
            config: V1ProjectVersion, the version config to download the artifacts for.
            path: str, the path where to persist the artifacts and assets.
        """
        if config.kind not in {
            V1ProjectVersionKind.MODEL,
            V1ProjectVersionKind.ARTIFACT,
        }:
            logger.info(
                "Skip artifacts download for version {} with kind {}.".format(
                    config.name, config.kind
                )
            )
            return

        meta_info = config.meta_info or {}
        run_info = meta_info.get("run", {})
        if not run_info:
            logger.info(
                "Skip artifacts download for version {} with kind {}. "
                "The version is not linked to any run.".format(config.name, config.kind)
            )
            return

        run_artifacts = [
            V1RunArtifact.from_dict(a) for a in meta_info.get("lineage", [])
        ]
        if not run_artifacts:
            logger.info(
                "Skip artifacts download for version {} with kind {}. "
                "The version is not linked to any artifacts.".format(
                    config.name, config.kind
                )
            )
            return

        run_project = run_info.get("project", self.project)
        run_uuid = run_info.get("uuid", config.run)

        from polyaxon.client.run import RunClient

        # Creating run client to download artifacts
        run_client = RunClient(owner=self.owner, project=run_project, run_uuid=run_uuid)
        for artifact_lineage in run_artifacts:
            logger.info(
                "Downloading artifact {} with kind {} and remote path {} to {}".format(
                    artifact_lineage.name,
                    artifact_lineage.kind,
                    artifact_lineage.path,
                    path,
                )
            )
            run_client.download_artifact_for_lineage(
                lineage=artifact_lineage, path_to=path
            )

    @client_handler(check_no_op=True, check_offline=True)
    def pull_version(
        self,
        kind: V1ProjectVersionKind,
        version: str,
        path: str,
        download_artifacts: bool = True,
    ):
        """Packages and downloads the version to a local path.

        This is a generic function based on the kind passed and pulls a:
            * component version
            * model version
            * artifact version

        Args:
            kind: V1ProjectVersionKind, kind of the project version.
            version: str, required, the version name/tag.
            path: str, optional, defaults to the offline root path,
                 path where to persist the metadata and artifacts.
            download_artifacts: bool, optional, to download the artifacts based on linked lineage.
        """
        path = ctx_paths.get_offline_path(
            entity_value=version, entity_kind=kind, path=path
        )
        delete_path(path)
        config = self.get_version(kind=kind, version=version)
        self.persist_version(config=config, path=path)
        if download_artifacts:
            self.download_artifacts_for_version(config=config, path=path)
        return path

    @client_handler(check_no_op=True, check_offline=True)
    def pull_component_version(
        self,
        version: str,
        path: str,
    ):
        """Packages and downloads the component version to a local path.

        Args:
            version: str, required, the version name/tag.
            path: str, local path where to persist the metadata and artifacts.
        """
        return self.pull_version(
            kind=V1ProjectVersionKind.COMPONENT,
            version=version,
            path=path,
            download_artifacts=False,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def pull_model_version(
        self,
        version: str,
        path: str,
        download_artifacts: bool = True,
    ):
        """Packages and downloads the model version to a local path.

        Args:
            version: str, required, the version name/tag.
            path: str, local path where to persist the metadata and artifacts.
            download_artifacts: bool, optional, to download the artifacts based on linked lineage.
        """
        return self.pull_version(
            kind=V1ProjectVersionKind.MODEL,
            version=version,
            path=path,
            download_artifacts=download_artifacts,
        )

    @client_handler(check_no_op=True, check_offline=True)
    def pull_artifact_version(
        self,
        version: str,
        path: str,
        download_artifacts: bool = True,
    ):
        """Packages and downloads the artifact version to a local path.

        Args:
            version: str, required, the version name/tag.
            path: str, local path where to persist the metadata and artifacts.
            download_artifacts: bool, optional, to download the artifacts based on linked lineage.
        """
        return self.pull_version(
            kind=V1ProjectVersionKind.ARTIFACT,
            version=version,
            path=path,
            download_artifacts=download_artifacts,
        )



import datetime

from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1Token(BaseResponseModel):
    uuid: Optional[UUIDStr]
    key: Optional[StrictStr]
    name: Optional[StrictStr]
    scopes: Optional[List[StrictStr]]
    services: Optional[List[StrictStr]]
    started_at: Optional[datetime.datetime]
    expires_at: Optional[datetime.datetime]
    created_at: Optional[datetime.datetime]
    updated_at: Optional[datetime.datetime]
    expiration: Optional[int]



import sys
import time

import click


@click.command()
@click.option("--uuid", help="The operation's run uuid.")
@click.option(
    "--kind",
    help="The operation kind.",
)
@click.option(
    "--max-retries",
    type=int,
    default=10,
    help="Number of times to retry the process.",
)
def wait(uuid: str, kind: str, max_retries: int):
    """Delete an s3 subpath."""
    from polyaxon import settings
    from polyaxon.k8s.executor.executor import Executor

    executor = Executor(namespace=settings.CLIENT_CONFIG.namespace, in_cluster=True)
    retry = 0
    while retry < max_retries:
        if retry:
            time.sleep(retry**2)
        try:
            k8s_operation = executor.get(run_uuid=uuid, run_kind=kind)
        except Exception:  # noqa
            k8s_operation = None
        if k8s_operation:
            retry += 1
        else:
            return

    sys.exit(1)



from typing import Dict, List, Optional, Union

from clipped.types.ref_or_obj import IntOrRef, RefField
from pydantic import Field, StrictStr, validator

from polyaxon.k8s import k8s_schemas, k8s_validation
from polyaxon.polyflow.environment import V1Environment
from polyaxon.polyflow.init import V1Init
from polyaxon.polyflow.run.resources import V1RunResources
from polyaxon.schemas.base import BaseSchemaModel


class V1RayReplica(BaseSchemaModel):
    """Ray replica is the specification for a Ray head or worker.

    Args:
        replicas: str, optional int
        min_replicas: str, int, optional
        max_replicas: str, int, optional
        ray_start_params: str, Dict[str, str], optional
        environment: [V1Environment](/docs/core/specification/environment/), optional
        connections: List[str], optional
        volumes: List[[Kubernetes Volume](https://kubernetes.io/docs/concepts/storage/volumes/)],
             optional
        init: List[[V1Init](/docs/core/specification/init/)], optional
        sidecars: List[[sidecar containers](/docs/core/specification/sidecars/)], optional
        container: [Kubernetes Container](https://kubernetes.io/docs/concepts/containers/)

    ## YAML usage

    ```yaml
    >>> head / workerName:
    >>>   replicas:
    >>>   minReplicas:
    >>>   maxReplicas:
    >>>   rayStartParams:
    >>>   environment:
    >>>   connections:
    >>>   volumes:
    >>>   init:
    >>>   sidecars:
    >>>   container:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.polyflow import V1Environment, V1Init, V1RayReplica
    >>> from polyaxon.k8s import k8s_schemas
    >>> replica = V1RayReplica(
    >>>     replicas=2,
    >>>     environment=V1Environment(...),
    >>>     init=[V1Init(...)],
    >>>     sidecars=[k8s_schemas.V1Container(...)],
    >>>     container=k8s_schemas.V1Container(...),
    >>> )
    ```

    ## Fields

    ### replicas

    The number of worker replica instances.

    ```yaml
    >>> executor:
    >>>   replicas: 2
    ```

    ### minReplicas

    The minimum number of worker replica instances.

    ```yaml
    >>> executor:
    >>>   minReplicas: 1
    ```

    ### maxReplicas

    The maximum number of worker replica instances.

    ```yaml
    >>> executor:
    >>>   minReplicas: 1
    ```

    ### rayStartParams

    The ray start parameters.

    ```yaml
    >>> head:
    >>>   rayStartParams:
    >>>     "dashboard-host": "0.0.0.0"
    >>>     "block": "true"
    >>>     "num-cpus": "0"
    ```

    ### environment

    Optional [environment section](/docs/core/specification/environment/),
    it provides a way to inject pod related information into the replica (executor/driver).

    ```yaml
    >>> worker:
    >>>   environment:
    >>>     labels:
    >>>        key1: "label1"
    >>>        key2: "label2"
    >>>      annotations:
    >>>        key1: "value1"
    >>>        key2: "value2"
    >>>      nodeSelector:
    >>>        node_label: node_value
    >>>      ...
    >>>  ...
    ```

    ### connections

    A list of [connection names](/docs/setup/connections/) to resolve for the job.

    <blockquote class="light">
    If you are referencing a connection it must be configured.
    All referenced connections will be checked:

     * If they are accessible in the context of the project of this run

     * If the user running the operation can have access to those connections
    </blockquote>

    After checks, the connections will be resolved and inject any volumes, secrets, configMaps,
    environment variables for your main container to function correctly.

    ```yaml
    >>> replica:
    >>>   connections: [connection1, connection2]
    ```

    ### init

    A list of [init handlers and containers](/docs/core/specification/init/)
    to resolve for the replica (executor/driver).

    <blockquote class="light">
    If you are referencing a connection it must be configured.
    All referenced connections will be checked:

     * If they are accessible in the context of the project of this run

     * If the user running the operation can have access to those connections
    </blockquote>

    ```yaml
    >>> replica:
    >>>   init:
    >>>     - artifacts:
    >>>         dirs: ["path/on/the/default/artifacts/store"]
    >>>     - connection: gcs-large-datasets
    >>>       artifacts:
    >>>         dirs: ["data"]
    >>>       container:
    >>>         resources:
    >>>           requests:
    >>>             memory: "256Mi"
    >>>             cpu: "500m"
    >>>     - container:
    >>>       name: myapp-container
    >>>       image: busybox:1.28
    >>>       command: ['sh', '-c', 'echo custom init container']
    ```

    ### sidecars


    A list of [sidecar containers](/docs/core/specification/sidecars/)
    that will be used as sidecars.

    ```yaml
    >>> replica:
    >>>   sidecars:
    >>>     - name: sidecar2
    >>>       image: busybox:1.28
    >>>       command: ['sh', '-c', 'echo sidecar2']
    >>>     - name: sidecar1
    >>>       image: busybox:1.28
    >>>       command: ['sh', '-c', 'echo sidecar1']
    >>>       resources:
    >>>         requests:
    >>>           memory: "128Mi"
    >>>           cpu: "500m"
    ```

    ### container

    The [main Kubernetes Container](https://kubernetes.io/docs/concepts/containers/)
    that will run your experiment training or data processing
    logic for the replica (executor/driver).

    ```yaml
    >>> replica:
    >>>   init:
    >>>     - connection: my-code-repo
    >>>   container:
    >>>     name: tensorflow:2.1
    >>>     command: ["python", "/plx-context/artifacts/my-code-repo/model.py"]
    ```
    """

    _IDENTIFIER = "replica"
    _SWAGGER_FIELDS = ["volumes", "sidecars", "container"]

    replicas: Optional[IntOrRef]
    min_replicas: Optional[IntOrRef] = Field(alias="minReplicas")
    max_replicas: Optional[IntOrRef] = Field(alias="maxReplicas")
    ray_start_params: Optional[Dict[str, str]] = Field(alias="rayStartParams")
    environment: Optional[Union[V1Environment, RefField]]
    connections: Optional[Union[List[StrictStr], RefField]]
    volumes: Optional[Union[List[k8s_schemas.V1Volume], RefField]]
    init: Optional[Union[List[V1Init], RefField]]
    sidecars: Optional[Union[List[k8s_schemas.V1Container], RefField]]
    container: Optional[Union[k8s_schemas.V1Container, RefField]]

    @validator("volumes", always=True, pre=True)
    def validate_volumes(cls, v):
        if not v:
            return v
        return [k8s_validation.validate_k8s_volume(vi) for vi in v]

    @validator("sidecars", always=True, pre=True)
    def validate_helper_containers(cls, v):
        if not v:
            return v
        return [k8s_validation.validate_k8s_container(vi) for vi in v]

    @validator("container", always=True, pre=True)
    def validate_container(cls, v):
        return k8s_validation.validate_k8s_container(v)

    def get_resources(self):
        resources = V1RunResources()
        for i in range(self.replicas or 1):
            resources += V1RunResources.from_container(self.container)

        return resources

    def get_all_containers(self):
        return [self.container] if self.container else []

    def get_all_connections(self):
        return self.connections or []

    def get_all_init(self):
        return self.init or []



from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, Any, List, Optional, Union
from typing_extensions import Annotated, Literal

from clipped.config.schema import skip_partial
from clipped.types.numbers import StrictIntOrFloat
from clipped.types.ref_or_obj import RefField
from pydantic import Field, StrictStr, root_validator, validator

from polyaxon import types
from polyaxon.polyflow.matrix.kinds import V1HPKind
from polyaxon.schemas.base import BaseSchemaModel

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator

try:
    import numpy as np
except (ImportError, ModuleNotFoundError):
    np = None


# pylint:disable=redefined-outer-name


def validate_pchoice(values):
    dists = [v for v in values if v]
    if sum(dists) > 1:
        raise ValueError("The distribution of different outcomes should sum to 1.")


class PChoice(tuple):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, (list, tuple)) and len(value) == 2:
            if isinstance(value[1], float) and 0 <= value[1] < 1:
                return value
        raise ValueError("This field expects a list of [value<Any>, dist<float>].")


def _validate_range(
    value: List, required_keys: List, keys: List, check_order: bool = True
):
    # TODO: Fix error message
    if len(value) < len(required_keys):
        raise ValueError(
            "This field expects a list of {} or {} elements, received {}.".format(
                len(required_keys), len(keys), len(value)
            )
        )
    # Check that lower value is smaller than higher value
    if check_order and value[0] >= value[1]:
        raise ValueError(
            "{key2} value must be strictly higher that {key1} value, "
            "received instead {key1}: {val1}, {key2}: {val2}".format(
                key1=required_keys[0],
                key2=required_keys[1],
                val1=value[0],
                val2=value[1],
            )
        )
    if len(required_keys) == 3 and not value[2]:
        raise ValueError(
            "{} must be > 0, received {}".format(required_keys[2], value[2])
        )


class BaseRange(BaseSchemaModel):
    _REQUIRED_KEYS = []
    _OPTIONAL_KEYS = []
    _CHECK_ORDER = True

    @root_validator
    def validate_range(cls, values):
        value = list(values.values())
        _validate_range(
            value,
            cls._REQUIRED_KEYS,
            cls._REQUIRED_KEYS + cls._OPTIONAL_KEYS,
            check_order=cls._CHECK_ORDER,
        )
        return values


class Range(BaseRange):
    start: Union[float]
    stop: Union[float]
    step: StrictIntOrFloat
    _REQUIRED_KEYS = ["start", "stop", "step"]


class RangeStr(StrictStr):
    _CLASS = Range

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value: str):
        if not isinstance(value, str) or ":" not in value:
            raise ValueError("This field expects a string of values separated by `:`.")
        keys = cls._CLASS._REQUIRED_KEYS + cls._CLASS._OPTIONAL_KEYS
        if value:
            value = value.split(":")
            return cls._CLASS(**dict(zip(keys, value)))
        raise ValueError(
            "`{}` requires `{}` or `{}` elements, received `{}`".format(
                cls._CLASS.__name__,
                len(cls._CLASS._REQUIRED_KEYS),
                len(keys),
                len(value),
            )
        )


class RangeList(list):
    _CLASS = Range

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value: List):
        if not isinstance(value, list):
            raise ValueError("This field expects a list of values.")
        keys = cls._CLASS._REQUIRED_KEYS + cls._CLASS._OPTIONAL_KEYS
        if value:
            return cls._CLASS(**dict(zip(keys, value)))
        raise ValueError(
            "`{}` requires `{}` or `{}` elements, received `{}`".format(
                cls._CLASS.__name__,
                len(cls._CLASS._REQUIRED_KEYS),
                len(keys),
                len(value),
            )
        )


class DateRange(Range):
    start: date
    stop: date
    step: int


class DateRangeList(RangeList):
    _CLASS = DateRange


class DateTimeRange(Range):
    start: datetime
    stop: datetime
    step: timedelta


class DateTimeRangeList(RangeList):
    _CLASS = DateTimeRange


class Space(BaseRange):
    start: Union[float]
    stop: Union[float]
    num: int
    _REQUIRED_KEYS = ["start", "stop", "num"]


class LinSpace(Space):
    pass


class LinSpaceList(RangeList):
    _CLASS = LinSpace


class LinSpaceStr(RangeStr):
    _CLASS = LinSpace


class GeomSpace(Space):
    pass


class GeomSpaceList(RangeList):
    _CLASS = GeomSpace


class GeomSpaceStr(RangeStr):
    _CLASS = GeomSpace


class LogSpace(Space):
    base: Optional[int]
    _OPTIONAL_KEYS = ["base"]


class LogSpaceList(RangeList):
    _CLASS = LogSpace


class LogSpaceStr(RangeStr):
    _CLASS = LogSpace


class Dist(BaseRange):
    low: Union[float]
    high: Union[float]
    size: Optional[int]
    _REQUIRED_KEYS = ["low", "high"]
    _OPTIONAL_KEYS = ["size"]
    _CHECK_ORDER = False


class QDist(BaseRange):
    low: Union[float]
    high: Union[float]
    q: StrictIntOrFloat
    size: Optional[int]
    _REQUIRED_KEYS = ["low", "high", "q"]
    _OPTIONAL_KEYS = ["size"]
    _CHECK_ORDER = False


class Uniform(Dist):
    pass


class UniformList(RangeList):
    _CLASS = Uniform


class UniformStr(RangeStr):
    _CLASS = Uniform


class QUniform(QDist):
    pass


class QUniformList(RangeList):
    _CLASS = QUniform


class QUniformStr(RangeStr):
    _CLASS = QUniform


class LogUniform(Dist):
    pass


class LogUniformList(RangeList):
    _CLASS = LogUniform


class LogUniformStr(RangeStr):
    _CLASS = LogUniform


class QLogUniform(QDist):
    pass


class QLogUniformList(RangeList):
    _CLASS = QLogUniform


class QLogUniformStr(RangeStr):
    _CLASS = QLogUniform


class Normal(BaseRange):
    loc: Union[float]
    scale: Union[float]
    size: Optional[int]
    _REQUIRED_KEYS = ["loc", "scale"]
    _OPTIONAL_KEYS = ["size"]
    _CHECK_ORDER = False


class NormalList(RangeList):
    _CLASS = Normal


class NormalStr(RangeStr):
    _CLASS = Normal


class QNormal(BaseRange):
    loc: Union[float]
    scale: Union[float]
    q: StrictIntOrFloat
    size: Optional[int]
    _REQUIRED_KEYS = ["loc", "scale", "q"]
    _OPTIONAL_KEYS = ["size"]
    _CHECK_ORDER = False


class QNormalList(RangeList):
    _CLASS = QNormal


class QNormalStr(RangeStr):
    _CLASS = QNormal


class LogNormal(Normal):
    pass


class LogNormalList(RangeList):
    _CLASS = LogNormal


class LogNormalStr(RangeStr):
    _CLASS = LogNormal


class QLogNormal(QNormal):
    pass


class QLogNormalList(RangeList):
    _CLASS = QLogNormal


class QLogNormalStr(RangeStr):
    _CLASS = QLogNormal


def validate_matrix(values):
    v = sum(map(lambda x: 1 if x else 0, values))
    if v == 0 or v > 1:
        raise ValueError(
            "Matrix element is not valid, one and only one option is required."
        )


class BaseHpParamConfig(BaseSchemaModel):
    _USE_DISCRIMINATOR = True

    @staticmethod
    def validate_io(io: "V1IO"):  # noqa
        if io.type not in ["int", "float"]:
            raise ValueError(
                "Param `{}` has a an input type `{}` "
                "and it does not correspond to hyper-param type `int or float`.".format(
                    io.name,
                    io.type,
                )
            )
        return True


class V1HpChoice(BaseHpParamConfig):
    """`Choice` picks a value from a of list values.

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: choice
    >>>     value: [1, 2, 3, 4, 5]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpChoice
    >>> param_test = V1HpChoice(value=[1, 2, 3, 4, 5])
    ```
    """

    _IDENTIFIER = V1HPKind.CHOICE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[List[Any], RefField]]

    @staticmethod
    def validate_io(io: "V1IO"):  # noqa
        return True

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return any(
            [
                v
                for v in self.value
                if not isinstance(v, (int, float, complex, np.integer, np.floating))
            ]
        )

    @property
    def is_uniform(self):
        return False


class V1HpPChoice(BaseHpParamConfig):
    """`PChoice` picks a value with a probability from a list of
    [(value, probability), (value, probability), ...].

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: pchoice
    >>>     value: [(1, 0.1), (2, 0.1), (3, 0.8)]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpPChoice
    >>> param_test = V1HpPChoice(value=[("A", 0.1), ("B", 0.1), ("C", 0.8)])
    ```
    """

    _IDENTIFIER = V1HPKind.PCHOICE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[List[PChoice], RefField]]

    @validator("value")
    @skip_partial
    def validate_value(cls, value):
        if value and isinstance(value, (list, tuple)):
            validate_pchoice(values=[v[1] for v in value if v])

        return value

    @staticmethod
    def validate_io(io: "V1IO"):  # noqa
        return True

    @property
    def is_distribution(self):
        return True

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpRange(BaseHpParamConfig):
    """`Range` picks a value from a generated list of values using `[start, stop, step]`,
    you can pass values in these forms:
      * [1, 10, 2]
      * {start: 1, stop: 10, step: 2}
      * '1:10:2'

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: range
    >>>     value: [1, 10, 2]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpRange
    >>> param_test = V1HpRange(value=[1, 10, 2])
    ```
    """

    _IDENTIFIER = V1HPKind.RANGE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[Range, RangeList, RangeStr, RefField]]

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpDateRange(BaseHpParamConfig):
    """`DateRange` picks a value from a generated list of values using `[start, stop, step]`,
    you can pass values in these forms:
      * `["2019-06-24", "2019-06-25", 3600 * 24]`
      * `{start: "2019-06-24 00:00", stop: "2019-06-28 00:00", step: 1}`

    **Step (frequency)**: represents a timedelta in days.

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: daterange
    >>>     value: ["2019-06-22", "2019-07-25", 1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpDateRange
    >>> param_test = V1HpDateRange(value=["2019-06-22", "2019-06-25", 2])
    ```
    """

    _IDENTIFIER = V1HPKind.DATERANGE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[DateRange, DateRangeList, RefField]]

    @staticmethod
    def validate_io(io: "V1IO"):  # noqa
        if io.type != types.DATE:
            raise ValueError(
                "Param `{}` has a an input type `{}` "
                "and it does not correspond to hyper-param type `date`.".format(
                    io.name,
                    io.type,
                )
            )
        return True

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpDateTimeRange(BaseHpParamConfig):
    """`DateTimeRange` picks a value from a generated list of values using `[start, stop, step]`,
    you can pass values in these forms:
      * `["2019-06-24T21:20:07+02:00", "2019-06-25T21:20:07+02:00", 3600]`
      * `{start: "2019-06-24 00:00", stop: "2019-06-28 00:00", step: 3600 * 4}`

    **Step (frequency)**: represents a timedelta in seconds.

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: datetimerange
    >>>     value: ["2019-06-22 21:00", "2019-06-25 21:00", 3600]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpDateTimeRange
    >>> param_test = V1HpDateTimeRange(value=["2019-06-22 21:00", "2019-06-25 21:00", 3600])
    ```
    """

    _IDENTIFIER = V1HPKind.DATETIMERANGE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[DateTimeRange, DateTimeRangeList, RefField]]

    @staticmethod
    def validate_io(io: "V1IO"):  # noqa
        if io.type != types.DATETIME:
            raise ValueError(
                "Param `{}` has a an input type `{}` "
                "and it does not correspond to hyper-param type `datetime`.".format(
                    io.name,
                    io.type,
                )
            )
        return True

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpLinSpace(BaseHpParamConfig):
    """`LinSpace` picks a value from a generated list of steps from start to stop spaced evenly
    on a linear scale `[start, stop, step]`, you can pass values in these forms:
      * [1, 10, 20]
      * {start: 1, stop: 10, num: 20}
      * '1:10:20'

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: linspace
    >>>     value: [1, 10, 20]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpLinSpace
    >>> param_test = V1HpLinSpace(value=[1, 10, 20])
    ```
    """

    _IDENTIFIER = V1HPKind.LINSPACE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[LinSpace, LinSpaceList, LinSpaceStr, RefField]]

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpLogSpace(BaseHpParamConfig):
    """`LogSpace` picks a value from a generated list of steps from start to stop spaced evenly
    on a log scale `[start, stop, step]` or `[start, stop, step, base]`,
    where `base` is optional with 10 as default value,
    you can pass values in these forms:
      * [1, 10, 20]
      * [1, 10, 20, 2]
      * {start: 1, stop: 10, num: 20}
      * {start: 1, stop: 10, num: 20, base: 2}
      * '1:10:20:2'

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: logspace
    >>>     value: [1, 10, 20]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpLogSpace
    >>> param_test = V1HpLinSpace(value=[1, 10, 20])
    ```
    """

    _IDENTIFIER = V1HPKind.LOGSPACE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[LogSpace, LogSpaceList, LogSpaceStr, RefField]]

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpGeomSpace(BaseHpParamConfig):
    """`GeomSpace` picks a value from a generated list of steps from start to stop spaced evenly
    on a geometric progression `[start, stop, step]`, you can pass values in these forms:
      * [1, 10, 20]
      * {start: 1, stop: 10, num: 20}
      * '1:10:20'

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: geomspace
    >>>     value: [1, 10, 20]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpGeomSpace
    >>> param_test = V1HpGeomSpace(value=[1, 10, 20])
    ```
    """

    _IDENTIFIER = V1HPKind.GEOMSPACE

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[GeomSpace, GeomSpaceList, GeomSpaceStr, RefField]]

    @property
    def is_distribution(self):
        return False

    @property
    def is_continuous(self):
        return False

    @property
    def is_discrete(self):
        return True

    @property
    def is_range(self):
        return True

    @property
    def is_categorical(self):
        return False

    @property
    def is_uniform(self):
        return False


class V1HpUniform(BaseHpParamConfig):
    """`Uniform` draws samples from a uniform distribution over the half-open
    interval `[low, high)`, you can pass values in these forms:
      * 0:1
      * [0, 1]
      * {'low': 0, 'high': 1}

    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: uniform
    >>>     value: [0, 1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpUniform
    >>> param_test = V1HpUniform(value=[0, 1])
    ```
    """

    _IDENTIFIER = V1HPKind.UNIFORM

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[Uniform, UniformList, UniformStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return True

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpQUniform(BaseHpParamConfig):
    """`QUniform` samples from a quantized uniform distribution over `[low, high]`
    (`round(uniform(low, high) / q) * q`),
    you can pass values in these forms:
      * 0:1:0.1
      * [0, 1, 0.1]
      * {'low': 0, 'high': 1, 'q': 0.1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: quniform
    >>>     value: [0, 1, 0.1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpQUniform
    >>> param_test = V1HpQUniform(value=[0, 1, 0.1])
    ```
    """

    _IDENTIFIER = "quniform"

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[QUniform, QUniformList, QUniformStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpLogUniform(BaseHpParamConfig):
    """`LogUniform` samples from a log uniform distribution over`[low, high]`,
    you can pass values in these forms:
      * 0:1
      * [0, 1]
      * {'low': 0, 'high': 1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: loguniform
    >>>     value: [0, 1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpLogUniform
    >>> param_test = V1HpLogUniform(value=[0, 1])
    ```
    """

    _IDENTIFIER = V1HPKind.LOGUNIFORM

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[LogUniform, LogUniformList, LogUniformStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpQLogUniform(BaseHpParamConfig):
    """`LogUniform` samples from a log uniform distribution over`[low, high]`,
    you can pass values in these forms:
      * 0:1:0.1
      * [0, 1, 0.1]
      * {'low': 0, 'high': 1, 'q': 0.1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: qloguniform
    >>>     value: [0, 1, 0.1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpQLogUniform
    >>> param_test = V1HpQLogUniform(value=[0, 1, 0.1])
    ```
    """

    _IDENTIFIER = V1HPKind.QLOGUNIFORM

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[QLogUniform, QLogUniformList, QLogUniformStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False

    @property
    def min(self):
        return None


class V1HpNormal(BaseHpParamConfig):
    """`Normal` draws random samples from a normal (Gaussian) distribution defined by
    `[loc, scale]`, you can pass values in these forms:
      * 0:1
      * [0, 1]
      * {'loc': 0, 'scale': 1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: normal
    >>>     value: [0, 1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpNormal
    >>> param_test = V1HpNormal(value=[0, 1])
    ```
    """

    _IDENTIFIER = V1HPKind.NORMAL

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[Normal, NormalList, NormalStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpQNormal(BaseHpParamConfig):
    """`QNormal` draws random samples from a quantized normal (Gaussian) distribution defined by
    `[loc, scale]`, you can pass values in these forms:
      * 0:1:0.1
      * [0, 1, 0.1]
      * {'loc': 0, 'scale': 1, 'q': 0.1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: qnormal
    >>>     value: [0, 1, 0.1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpQNormal
    >>> param_test = V1HpNormal(value=[0, 1, 0.1])
    ```
    """

    _IDENTIFIER = V1HPKind.QNORMAL

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[QNormal, QNormalList, QNormalStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpLogNormal(BaseHpParamConfig):
    """`LogNormal` draws random samples from a log normal (Gaussian) distribution defined by
    `[loc, scale]`, you can pass values in these forms:
      * 0:1
      * [0, 1]
      * {'loc': 0, 'scale': 1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: lognormal
    >>>     value: [0, 1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpLogNormal
    >>> param_test = V1HpLogNormal(value=[0, 1])
    ```
    """

    _IDENTIFIER = V1HPKind.LOGNORMAL

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[LogNormal, LogNormalList, LogNormalStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


class V1HpQLogNormal(BaseHpParamConfig):
    """`QLogNormal` draws random samples from a log normal (Gaussian) distribution defined by
    `[loc, scale]`, you can pass values in these forms:
      * 0:1:0.1
      * [0, 1, 0.1]
      * {'loc': 0, 'scale': 1, 'q': 0.1}


    ```yaml
    >>> params:
    >>>   paramTest:
    >>>     kind: qlognormal
    >>>     value: [0, 1, 0.1]
    ```

    ```python
    >>> from polyaxon.polyflow import V1HpQLogNormal
    >>> param_test = V1HpQLogNormal(value=[0, 1])
    ```
    """

    _IDENTIFIER = V1HPKind.QLOGNORMAL

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    value: Optional[Union[QLogNormal, QLogNormalList, QLogNormalStr, RefField]]

    @property
    def is_distribution(self):
        return True

    @property
    def is_uniform(self):
        return False

    @property
    def is_continuous(self):
        return True

    @property
    def is_discrete(self):
        return False

    @property
    def is_range(self):
        return False

    @property
    def is_categorical(self):
        return False


V1HpParam = Annotated[
    Union[
        V1HpChoice,
        V1HpPChoice,
        V1HpRange,
        V1HpDateRange,
        V1HpDateTimeRange,
        V1HpLinSpace,
        V1HpLogSpace,
        V1HpGeomSpace,
        V1HpUniform,
        V1HpQUniform,
        V1HpLogUniform,
        V1HpQLogUniform,
        V1HpNormal,
        V1HpQNormal,
        V1HpLogNormal,
        V1HpQLogNormal,
    ],
    Field(discriminator="kind"),
]



from typing import Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1SettingsCatalog(BaseResponseModel):
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]



from typing import List

from polyaxon.contexts import paths as ctx_paths


def get_artifacts_store_args(artifacts_path: str, clean: bool) -> str:
    get_or_create = 'if [ ! -d "{dir}" ]; then mkdir -m 0777 -p {dir}; fi;'.format(
        dir=artifacts_path
    )
    delete_dir = (
        'if [ -d {path} ] && [ "$(ls -A {path})" ]; '
        "then rm -R {path}/*; fi;".format(path=artifacts_path)
    )
    if clean:
        return "{} {}".format(get_or_create, delete_dir)
    return "{}".format(get_or_create)


def init_artifact_context_args(run_path: str) -> List[str]:
    return [
        'if [ ! -d "{dir}" ]; then mkdir -m 0777 -p {dir}; fi;'.format(
            dir=ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(run_path)
        ),
        'if [ ! -d "{dir}" ]; then mkdir -m 0777 -p {dir}; fi;'.format(
            dir=ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(run_path) + "/outputs"
        ),
    ]



from typing import Optional

from clipped.types.email import EmailStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class UserConfig(BaseSchemaModel):
    _IDENTIFIER = "user"

    username: StrictStr
    email: Optional[EmailStr]
    name: Optional[StrictStr]
    theme: Optional[int]



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from clipped.types.ref_or_obj import IntOrRef, RefField
from pydantic import Field, PositiveInt, validator

from polyaxon.polyflow.early_stopping import V1EarlyStopping
from polyaxon.polyflow.matrix.base import BaseSearchConfig
from polyaxon.polyflow.matrix.kinds import V1MatrixKind
from polyaxon.polyflow.matrix.params import V1HpParam


class V1RandomSearch(BaseSearchConfig):
    """Random search creates a number of unique experiments by sampling randomly
    from a search space.
    Random search is a competitive method for black-box parameter tuning in machine learning.

    Random search requires a parameter `numRuns`,
    this is essential because Polyaxon needs to know how many experiments to sample.

    Args:
        kind: str, should be equal `grid`
        params: List[Dict[str, [params](/docs/automation/optimization-engine/params/)]]
        concurrency: int, optional
        num_runs: int, optional
        seed: int, optional
        early_stopping: List[[EarlyStopping](/docs/automation/helpers/early-stopping)], optional

    ## YAML usage

    ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   concurrency:
    >>>   params:
    >>>   numRuns:
    >>>   seed:
    >>>   earlyStopping:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.polyflow import (
    >>>     V1RandomSearch, V1HpLogSpace, V1HpUniform, V1FailureEarlyStopping, V1MetricEarlyStopping
    >>> )
    >>> matrix = V1RandomSearch(
    >>>   num_runs=20,
    >>>   concurrency=2,
    >>>   seed=23,
    >>>   params={"param1": V1HpLogSpace(...), "param2": V1HpUniform(...), ... },
    >>>   early_stopping=[V1FailureEarlyStopping(...), V1MetricEarlyStopping(...)]
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this matrix is a random search.

    If you are using the python client to create the mapping,
    this field is not required and is set by default.

    ```yaml
    >>> matrix:
    >>>   kind: random
    ```

    ### concurrency

    An optional value to set the number of concurrent operations.

    <blockquote class="light">
    This value only makes sense if less or equal to the total number of possible runs.
    </blockquote>

    ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   concurrency: 2
    ```

    For more details about concurrency management,
    please check the [concurrency section](/docs/automation/helpers/concurrency/).

    ### params

    A dictionary of `key -> value generator`
    to generate the parameters.

    To learn about all possible
    [params generators](/docs/automation/optimization-engine/params/).

    > The parameters generated will be validated against
    > the component's inputs/outputs definition to check that the values
    > can be passed and have valid types.

    ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   params:
    >>>     param1:
    >>>        kind: ...
    >>>        value: ...
    >>>     param2:
    >>>        kind: ...
    >>>        value: ...
    ```

    ### numRuns

    Maximum number of runs to start based on the search space defined.

    ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   numRuns: 5
    ```

    ### seed

    Since this algorithm uses random generators,
    if you want to control the seed for the random generator, you can pass a seed.

     ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   seed: 523
    ```

    ### earlyStopping

    A list of early stopping conditions to check for terminating
    all operations managed by the pipeline.
    If one of the early stopping conditions is met,
    a signal will be sent to terminate all running and pending operations.

    ```yaml
    >>> matrix:
    >>>   kind: random
    >>>   earlyStopping: ...
    ```

    For more details please check the
    [early stopping section](/docs/automation/helpers/early-stopping/).

    ## Example

    In this example the random search algorithm will try 20 unique experiments based on the
    search space defined in the params subsection.


    ```yaml
    >>> version: 1.1
    >>> kind: operation
    >>> matrix:
    >>>   kind: random
    >>>   concurrency: 10
    >>>   numRuns: 20
    >>>   params:
    >>>     lr:
    >>>       kind: logspace
    >>>       value: 0.01:0.1:5
    >>>     dropout:
    >>>       kind: choice
    >>>       value: [0.2, 0.5]
    >>>    activation:
    >>>       kind: pchoice
    >>>       value: [[elu, 0.1], [relu, 0.2], [sigmoid, 0.7]]
    >>>    early_stopping:
    >>>      - metric: accuracy
    >>>        value: 0.9
    >>>        optimization: maximize
    >>>      - metric: loss
    >>>        value: 0.05
    >>>        optimization: minimize
    >>> component:
    >>>   inputs:
    >>>     - name: batch_size
    >>>       type: int
    >>>       isOptional: true
    >>>       value: 128
    >>>     - name: lr
    >>>       type: float
    >>>     - name: dropout
    >>>       type: float
    >>>     - name: activation
    >>>       type: str
    >>>   container:
    >>>     image: image:latest
    >>>     command: [python3, train.py]
    >>>     args: [
    >>>         "--batch-size={{ batch_size }}",
    >>>         "--lr={{ lr }}",
    >>>         "--dropout={{ dropout }}",
    >>>         "--activation={{ activation }}"
    >>>     ]
    ```
    """

    _IDENTIFIER = V1MatrixKind.RANDOM

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    params: Union[Dict[str, V1HpParam], RefField]
    num_runs: Union[PositiveInt, RefField] = Field(alias="numRuns")
    seed: Optional[IntOrRef]
    concurrency: Optional[Union[PositiveInt, RefField]]
    early_stopping: Optional[Union[List[V1EarlyStopping], RefField]] = Field(
        alias="earlyStopping"
    )

    @validator("num_runs", "concurrency", pre=True)
    def check_values(cls, v, field):
        if v and v < 1:
            raise ValueError(f"{field} must be greater than 1, received `{v}` instead.")
        return v



from typing import List, Union

from clipped.utils.lists import to_list
from clipped.utils.paths import check_or_create_path
from fsspec import AbstractFileSystem

from polyaxon.logger import logger


def download_file_or_dir(
    fs: AbstractFileSystem,
    path_from: str,
    path_to: str,
    is_file: bool,
    check_path: bool,
):
    if check_path:
        is_file = fs.isfile(path_from)
    check_or_create_path(path_to, is_dir=not is_file)
    if not is_file:
        path_to = path_to.rstrip("/") + "/"
        path_from = path_from.rstrip("/") + "/"
    fs.download(rpath=path_from, lpath=path_to, recursive=not is_file)


def delete_file_or_dir(
    fs: AbstractFileSystem,
    subpath: Union[str, List[str]],
    is_file: bool = False,
):
    subpath = to_list(subpath, check_none=True)
    for sp in subpath:
        try:
            fs.delete(path=sp, recursive=not is_file)
        except Exception as e:
            logger.info("Could not delete path %s\nError: %s", sp, e)



from clipped.utils.units import to_cpu_value, to_memory_bytes

from polyaxon.constants.globals import UNKNOWN


class NodeLifeCycle:
    UNKNOWN = UNKNOWN
    READY = "ready"
    NOT_READY = "notReady"
    DELETED = "deleted"

    CHOICES = (
        (UNKNOWN, UNKNOWN),
        (READY, READY),
        (NOT_READY, NOT_READY),
        (DELETED, DELETED),
    )


class NodeParser:
    @staticmethod
    def get_status(node) -> str:
        status = [c.status for c in node.status.conditions if c.type == "Ready"][0]
        if status == "True":
            return NodeLifeCycle.READY
        if status == "False":
            return NodeLifeCycle.NOT_READY
        return NodeLifeCycle.UNKNOWN

    @staticmethod
    def get_n_gpus(node) -> int:
        if "gpu" not in node.status.allocatable:
            return 0
        return int(node.status.allocatable.get("nvidia.com/gpu", 0))

    @staticmethod
    def get_cpu(node) -> float:
        cpu = node.status.allocatable["cpu"]
        return to_cpu_value(cpu)

    @classmethod
    def get_memory(cls, node) -> int:
        return to_memory_bytes(node.status.allocatable["memory"])

    @staticmethod
    def get_runtime(node) -> str:
        return node.status.node_info.container_runtime_version

    @staticmethod
    def get_schedulable_state(node) -> bool:
        return not node.spec.unschedulable

    @staticmethod
    def get_addresses(node) -> str:
        return node.status.addresses



from typing import Union
from typing_extensions import Annotated

from pydantic import Field

from polyaxon.polyflow.schedules.cron import V1CronSchedule
from polyaxon.polyflow.schedules.datetime import V1DateTimeSchedule
from polyaxon.polyflow.schedules.interval import V1IntervalSchedule
from polyaxon.polyflow.schedules.kinds import V1ScheduleKind

V1Schedule = Annotated[
    Union[V1IntervalSchedule, V1CronSchedule, V1DateTimeSchedule],
    Field(discriminator="kind"),
]


class ScheduleMixin:
    def get_schedule_kind(self):
        raise NotImplementedError

    @property
    def has_interval_schedule(self):
        return self.get_schedule_kind() == V1IntervalSchedule._IDENTIFIER

    @property
    def has_cron_schedule(self):
        return self.get_schedule_kind() == V1CronSchedule._IDENTIFIER

    @property
    def has_datetime_schedule(self):
        return self.get_schedule_kind() == V1DateTimeSchedule._IDENTIFIER



import sys
import time

import click

from polyaxon.exceptions import PolyaxonAgentError
from polyaxon.logger import logger


@click.group()
def agent():
    pass


@agent.command()
@click.option(
    "--kind",
    type=str,
    default="k8s",
    help="The agent executor backend to use.",
)
@click.option(
    "--sleep-interval",
    type=int,
    help="Sleep interval between fetches (Applied only to base agent).",
)
@click.option(
    "--max-retries",
    type=int,
    default=3,
    help="Number of times to retry the process.",
)
def start(kind, max_retries, sleep_interval):
    from polyaxon import settings
    from polyaxon.env_vars.getters import get_agent_info
    from polyaxon.runner.kinds import RunnerKind

    kind = kind or RunnerKind.K8S

    if kind == RunnerKind.K8S:
        from polyaxon.k8s.agent import Agent
    elif kind == RunnerKind.DOCKER:
        from polyaxon.docker.agent import Agent
    else:
        logger.error("Received an unsupported agent kind: `{}`".format(kind))
        sys.exit(1)

    settings.CLIENT_CONFIG.set_agent_header()
    owner, agent_uuid = None, None
    try:
        owner, agent_uuid = get_agent_info()
        logger.info("Using agent with info: {}, {}".format(owner, agent_uuid))
    except PolyaxonAgentError:
        logger.info("Using base agent")

    retry = 0
    while retry < max_retries:
        if retry:
            time.sleep(5 * retry)
        try:
            Agent(
                owner=owner, agent_uuid=agent_uuid, sleep_interval=sleep_interval
            ).start()
            return
        except Exception as e:
            logger.warning("Polyaxon agent retrying, error %s", e)
            retry += 1


@agent.command()
@click.option(
    "--health-interval",
    type=int,
    help="Health interval between checks.",
)
def healthz(health_interval):
    from polyaxon.runner.agent import BaseAgent

    if not BaseAgent.pong(interval=health_interval):
        logger.warning("Polyaxon agent is not healthy!")
        sys.exit(1)



# import apis into sdk package
from polyaxon.sdk.api.agents_v1_api import AgentsV1Api
from polyaxon.sdk.api.artifacts_stores_v1_api import ArtifactsStoresV1Api
from polyaxon.sdk.api.auth_v1_api import AuthV1Api
from polyaxon.sdk.api.connections_v1_api import ConnectionsV1Api
from polyaxon.sdk.api.dashboards_v1_api import DashboardsV1Api
from polyaxon.sdk.api.organizations_v1_api import OrganizationsV1Api
from polyaxon.sdk.api.presets_v1_api import PresetsV1Api
from polyaxon.sdk.api.project_dashboards_v1_api import ProjectDashboardsV1Api
from polyaxon.sdk.api.project_searches_v1_api import ProjectSearchesV1Api
from polyaxon.sdk.api.projects_v1_api import ProjectsV1Api
from polyaxon.sdk.api.queues_v1_api import QueuesV1Api
from polyaxon.sdk.api.runs_v1_api import RunsV1Api
from polyaxon.sdk.api.searches_v1_api import SearchesV1Api
from polyaxon.sdk.api.service_accounts_v1_api import ServiceAccountsV1Api
from polyaxon.sdk.api.tags_v1_api import TagsV1Api
from polyaxon.sdk.api.teams_v1_api import TeamsV1Api
from polyaxon.sdk.api.users_v1_api import UsersV1Api
from polyaxon.sdk.api.versions_v1_api import VersionsV1Api

# import ApiClient
from polyaxon.sdk.async_client.api_client import AsyncApiClient
from polyaxon.sdk.configuration import Configuration
from polyaxon.sdk.exceptions import (
    ApiException,
    ApiKeyError,
    ApiTypeError,
    ApiValueError,
    OpenApiException,
)
from polyaxon.sdk.sync_client.api_client import ApiClient



import sys

import click

from clipped.formatting import Printer
from clipped.utils.dicts import dict_to_tabulate

from polyaxon import settings
from polyaxon.cli.errors import handle_cli_error
from polyaxon.cli.session import set_versions_config
from polyaxon.logger import clean_outputs, logger
from polyaxon.managers.cli import CliConfigManager
from polyaxon.managers.client import ClientConfigManager
from polyaxon.managers.home import HomeConfigManager
from polyaxon.managers.user import UserConfigManager


def validate_options(ctx, param, value):
    possible_values = ["verbose", "host"]
    if value and value not in possible_values:
        raise click.BadParameter(
            "Value `{}` is not supported, must one of the value {}".format(
                value, possible_values
            )
        )
    return value


@click.group(invoke_without_command=True)
@click.option(
    "--list",
    "-l",
    "_list",
    is_flag=True,
    help="Deprecated, please use `polyaxon config show`.",
)
@clean_outputs
def config(_list):  # pylint:disable=redefined-builtin
    """Set and get the global configurations."""
    if _list:
        Printer.warning(
            "`polyaxon config -l` is deprecated, please use `polyaxon config show`!"
        )


@config.command()
@clean_outputs
def show():
    """Show the current cli, client, and user configs."""
    _config = HomeConfigManager.get_config_or_default()
    Printer.heading(
        "In addition to environment variables, global configs will be loaded from:"
    )
    Printer.dict_tabulate(_config.to_dict())
    _config = ClientConfigManager.get_config_or_default()
    Printer.heading("Client config:")
    Printer.dict_tabulate(_config.to_dict())
    _config = CliConfigManager.get_config_or_default()
    if _config:
        Printer.heading("CLI config:")
        if _config.current_version:
            Printer.print("Version {}".format(_config.current_version))
        else:
            Printer.warning("This cli is not configured.")
        if _config.installation:
            config_installation = dict_to_tabulate(
                _config.installation,
                humanize_values=True,
                exclude_attrs=["hmac", "auth", "host"],
            )
            Printer.heading("Platform config:")
            Printer.dict_tabulate(config_installation)
        else:
            Printer.warning("This cli is not connected to a Polyaxon Host.")
    _config = UserConfigManager.get_config_or_default()
    if _config:
        Printer.heading("User config:")
        config_user = dict_to_tabulate(
            _config.to_dict(),
            humanize_values=True,
            exclude_attrs=["theme"],
        )
        Printer.dict_tabulate(config_user)


@config.command()
@click.argument("keys", type=str, nargs=-1)
@clean_outputs
def get(keys):
    """Get the specific keys from the global configuration.

    Examples:

    \b
    $ polyaxon config get host verify-ssl
    """
    _config = ClientConfigManager.get_config_or_default()

    if not keys:
        return

    print_values = {}
    for key in keys:
        key = key.replace("-", "_")
        if hasattr(_config, key):
            print_values[key] = getattr(_config, key)
        else:
            Printer.print("Key `{}` is not recognised.".format(key))

    Printer.dict_tabulate(print_values)


@config.command()
@click.option("--debug", type=bool, help="To set the verbosity of the client.")
@click.option("--host", type=str, help="To set the server endpoint.")
@click.option(
    "--no-api",
    type=bool,
    help="To disable any API call.",
)
@click.option(
    "--verify-ssl",
    type=bool,
    help="To set whether or not to verify the SSL certificate.",
)
@click.option(
    "--home",
    type=click.Path(exists=True),
    help="To set POLYAXON_HOME to specify the context where the CLI/Client reads/writes global configuration.",
)
@click.option(
    "--disable-errors-reporting",
    type=bool,
    help="To set the disable errors reporting.",
)
@click.option(
    "--no-purge",
    is_flag=True,
    default=False,
    help="To reconfigure the host without purging auth and other config options.",
)
@clean_outputs
def set(**kwargs):  # pylint:disable=redefined-builtin
    """Set the global config values.

    Examples:

    \b
    $ polyaxon config set --host=localhost
    """
    no_purge = kwargs.pop("no_purge", None)
    if kwargs.get("home") is not None:
        try:
            _config = HomeConfigManager.get_config_or_default()
        except Exception as e:
            logger.debug(
                "Home configuration could not be loaded.\n"
                "Error: %s\n"
                "Purging home configuration and resetting values.",
                e,
            )
            logger.debug()
            HomeConfigManager.purge()
            _config = HomeConfigManager.get_config_or_default()
        setattr(_config, "path", kwargs.pop("home", None))
        HomeConfigManager.set_config(_config)

    from polyaxon.managers.auth import AuthConfigManager

    try:
        _config = ClientConfigManager.get_config_or_default()
    except Exception as e:
        handle_cli_error(e, message="Polyaxon load configuration.")
        Printer.heading("You can reset your config by running: `polyaxon config purge`")
        sys.exit(1)

    should_purge = False
    for key, value in kwargs.items():
        if key == "host" and value is not None:
            should_purge = True
        if value is not None:
            setattr(_config, key, value)

    ClientConfigManager.set_config(_config)
    Printer.success("Config was updated.")
    # Reset cli config
    CliConfigManager.purge()
    if should_purge and not no_purge:
        AuthConfigManager.purge()
        UserConfigManager.purge()
        settings.set_client_config()
        set_versions_config()


@config.command()
@click.option(
    "--cache-only",
    is_flag=True,
    help="To purge the cache only.",
)
@clean_outputs
def purge(cache_only):
    """Purge the global config values."""
    from polyaxon.managers.auth import AuthConfigManager
    from polyaxon.managers.project import ProjectConfigManager
    from polyaxon.managers.run import RunConfigManager

    if not cache_only:
        ClientConfigManager.purge()
        CliConfigManager.purge()
        AuthConfigManager.purge()
        UserConfigManager.purge()
    ProjectConfigManager.purge()
    RunConfigManager.purge()
    Printer.success("Configs was removed.")



from typing import Optional

from polyaxon.schemas.base import BaseSchemaModel


class OperatorsConfig(BaseSchemaModel):
    tfjob: Optional[bool]
    pytorchjob: Optional[bool]
    paddlejob: Optional[bool]
    mpijob: Optional[bool]
    mxjob: Optional[bool]
    xgbjob: Optional[bool]
    sparkjob: Optional[bool]
    daskjob: Optional[bool]
    rayjob: Optional[bool]
